/*
 * Copyright 2011 Raster Software Vigo
 *
 * This file is part of ttos
 *
 * ttos is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * ttos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * bitmaps.c
 *
 *  Created on: 20/09/2010
 *      Author: raster
 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <endian.h>
#include <stdlib.h>
#include <stdio.h>

#include "bitmaps.h"
#include "global_vars.h"

/**
 * Loads a bitmap in memory and returns a pointer to a TTOS_Surface; if the bitmap was already loaded by another client,
 * it will be retrieved from memory, to reuse it.
 *
 * @param filename Path and filename to the bitmap (in TGA format)
 *
 * @return a TTOS_Surface with the bitmap, or NULL on error
 */

TTOS_Surface *bitmaps_load(Cstring filename) {

	struct _bitmap_container *container;
	struct minibus_list_element *element;

	Cstring final_path,base_path,tmp_path;

	if (!filename) {
		return NULL;
	}

	tmp_path=getenv("TTOS_BASE_PATH");
	if ((!tmp_path)||(!tmp_path[0])) {
		final_path=minibus_path_join("pics/",filename);
	} else {
		base_path=minibus_path_join(tmp_path,"pics");
		final_path=minibus_path_join(base_path,filename);
		free(base_path);
	}

	for(element=bitmaps.list;element!=NULL;element=element->next) {
		container=element->data;
		if (!strcmp(final_path,container->filename)) { // the bitmap is already loaded
			free(final_path);
			minibus_list_incref(element); // increment the reference count
			return (container->bitmap); // and return it
		}
	}

	// if the code reachs this, the bitmap hasn't been loaded

	container=malloc(sizeof(struct _bitmap_container));
	if (!container) {
		free(final_path);
		return NULL;
	}
	container->filename=strdup(final_path);
	container->bitmap=bitmaps_load_TGA(final_path);
	free(final_path);
	if (!container->bitmap) {
		free(container->filename);
		free(container);
		return(NULL);
	}
	minibus_list_add(&bitmaps,container);
	return(container->bitmap);
}

/**
 * Frees a _bitmap_container structure
 *
 * @param container The structure to free
 *
 */

void bitmaps_free_container(void *element) {

	struct _bitmap_container *container=element;

	if (!container) {
		return;
	}
	if (container->filename) {
		free(container->filename);
	}
	if (container->bitmap) {
		blit_destroy_surface(container->bitmap);
	}
	free(container);
}

/**
 * Frees a reference to a bitmap, by decrementing its reference counter in the bitmap list. The bitmap is removed from
 * the list only when all the clients using it have freed it (i.e. its reference count reaches zero)
 *
 * @param bitmap The bitmap to be freed
 */

void bitmaps_unload(TTOS_Surface *bitmap) {

	struct _bitmap_container *container;
	struct minibus_list_element *element;

	for(element=bitmaps.list;element!=NULL;element=element->next) {
		container=element->data;
		if (container->bitmap==bitmap) {
			minibus_list_decref(element);
			break;
		}
	}
}

/**
 * Loads a TGA image file into a surface, and returns it in a surface.
 *
 * @param filename The TGA filename
 *
 * @return A surface with the picture, or NULL if there's an error
 */

TTOS_Surface *bitmaps_load_TGA(Cstring filename) {

	int fd;
	uint8_t header[256],*rgba,*o_pointer,*a_pointer;
	uint16_t r,g,b,a,*d_pointer;
	bool has_alpha,is_rle;
	TTOS_Surface *surface;
	int32_t jump,d_line,o_jump,base_jump,rle_counter;
	uint32_t x,y,w,h,Bpp,tmp;
	uint32_t counter,max;

	fd=open(filename,O_RDONLY);

	if(fd<0) {
		return NULL;
	}

	if (18!=read(fd,header,18)) {
		close(fd);
		return NULL;
	}

	if (header[1]) {
		close(fd);
		return NULL; // No color maps allowed
	}

	if ((header[2]!=2)&&(header[2]!=10)) {
		close(fd);
		return NULL; // only TrueColor pictures allowed
	}

	if ((header[16]!=24)&&(header[16]!=32)) {
		close(fd);
		return NULL; // Only 24 (no alpha) or 32 (with alpha) bpps allowed
	}

	Bpp = header[16]==24 ? 3 : 4;

	if (header[17]&0x07) { // only 0 or 8 bits for alpha channel
		close(fd);
		return NULL;
	}

	has_alpha = (header[17]&0x08) ? 1 : 0;

	w=(int32_t)(header[12])+256*(int32_t)(header[13]);
	h=(int32_t)(header[14])+256*(int32_t)(header[15]);
	surface=blit_create_surface(w,h,has_alpha);
	if (!surface)
		return (NULL);
	d_pointer=surface->pixels;
	a_pointer=surface->alpha;

	switch(header[17]&0x30) {
	case 0x20: // origin up-left
		jump=1;
		d_line=0;
		base_jump=0;
		break;
	case 0x10: // origin down-right
		base_jump=w*h-1;
		jump=-1;
		d_line=0;
		break;
	case 0x00: // origin down-left
		base_jump=w*(h-1);
		jump=1;
		d_line=-w*2;
		break;
	case 0x30: // origin up-right
		base_jump=w-1;
		jump=-1;
		d_line=w*2;
		break;
	default:
		break;
	}

	if(header[0]) {
		read(fd,header,(uint32_t)header[0]);
	}

	if (header[2]==2) { // uncompressed RGB
		rgba=malloc(Bpp*w*h);
		if (!rgba) {
			close(fd);
			return (NULL);
		}
		read(fd,rgba,Bpp*w*h);
		is_rle=false;
		o_jump=0;
	} else { // compressed RGB
		rgba=malloc(2+Bpp*128);
		if (!rgba) {
			close(fd);
			return (NULL);
		}
		read(fd,rgba,1);
		rle_counter=1;
		o_jump=0;
		is_rle=true;
	}

	d_pointer+=base_jump;
	if (has_alpha) {
		a_pointer+=base_jump;
	}
	o_pointer=rgba;
	for(y=0;y<h;y++) {
		for(x=0;x<w;x++) {
			if (is_rle) {
				rle_counter--;
				o_pointer+=o_jump;
				if (!rle_counter) {
					if (*rgba<128) {
						rle_counter=1+(int32_t)(*rgba);
						o_jump=0;
						tmp=1+Bpp*rle_counter;
					} else {
						rle_counter=1+(int32_t)((*rgba)-128);
						o_jump=-Bpp;
						tmp=1+Bpp;
					}
					read(fd,rgba+1,tmp);
					*rgba=rgba[tmp];
					o_pointer=rgba+1;
				}
			}
			b=(uint16_t)*(o_pointer++);
			g=(uint16_t)*(o_pointer++);
			r=(uint16_t)*(o_pointer++);
			if (has_alpha) {
				a=(uint16_t)*(o_pointer++);
				if (a!=255) {
					b=((b*a)>>8); /* R, G and B values must be premultiplied to ease calculations */
					g=((g*a)>>8);
					r=((r*a)>>8);
				}
				a=255-a;
			} else {
				a=0;
			}
			*(d_pointer)=((r<<8)&0xF800)|((g<<3)&0x07E0)|((b>>3)&0x001F);
			if (has_alpha) {
				*a_pointer=(uint8_t)a;
				a_pointer+=jump;
			}
			d_pointer+=jump;
		}
		if (has_alpha)
			a_pointer+=d_line;
		d_pointer+=d_line;
	}
	free(rgba);
	close(fd);
	if (!surface->alpha) {
		return (surface);
	}

	a_pointer=surface->alpha;
	max=surface->w*surface->h;
	for(counter=0;counter<max;counter++) {
		if (*(a_pointer++)!=255) {
			return (surface);
		}
	}
	free(surface->alpha);
	surface->alpha=NULL;
	return(surface);
}
