/*
 * Copyright 2011 Raster Software Vigo
 *
 * This file is part of ttos
 *
 * ttos is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * ttos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * blit.c
 *
 *  Created on: 08/09/2010
 *      Author: raster
 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <endian.h>
#include <stdlib.h>

#include "ttos_lib/ttos_structs.h"
#include "blit.h"
#include "sharemem.h"

/**
 * Creates an empty surface
 *
 * @param w Width
 * @param h Height
 * @param alpha If true, will have an alpha channel; if false, won't
 *
 * @return A TTOS_Surface if successful; NULL on error
 *
 */

TTOS_Surface *blit_create_surface(uint32_t w, uint32_t h,bool alpha) {

	TTOS_Surface *surface;

	surface=malloc(sizeof(TTOS_Surface));
	if (!surface)
		return (NULL);
	surface->w=w;
	surface->h=h;
	surface->pixels=malloc(2*w*h);
	if (!surface->pixels) {
		free(surface);
		return (NULL);
	}
	if (alpha) {
		surface->alpha=malloc(w*h);
		if (!surface->alpha) {
			free(surface->pixels);
			free(surface);
			return (NULL);
		}
	} else {
		surface->alpha=NULL;
	}
	surface->uses_shared_mem=false;
	return (surface);
}

/**
 * Creates an empty surface using a shared memory object
 *
 * @param sharemem The shared memory object
 * @param alpha If true, will have an alpha channel; if false, won't
 *
 * @return A TTOS_Surface if successful; NULL on error
 *
 */

TTOS_Surface *blit_create_surface_from_shared(struct _shared_mem *sharemem, bool alpha) {

	TTOS_Surface *surface;
	uint8_t *pointer;

	pointer=sharemem->pixels;

	surface=malloc(sizeof(TTOS_Surface));
	if (!surface)
		return (NULL);
	surface->w=sharemem->width;
	surface->h=sharemem->height;
	surface->pixels=sharemem->pixels;

	if ((alpha)&&(sharemem->has_alpha)) {
		surface->alpha=pointer+(2*surface->w*surface->h);
	} else {
		surface->alpha=NULL;
	}
	surface->uses_shared_mem=true;
	return (surface);
}


/**
 * Frees a surface
 *
 * @param surface The surface to free
 *
 */

void blit_destroy_surface(TTOS_Surface *surface) {

	if (surface) {
		if (!surface->uses_shared_mem) {
			if (surface->pixels)
				free(surface->pixels);
			if (surface->alpha)
				free(surface->alpha);
		}
		free(surface);
	}
}


/**
 * Blends a piece of a surface into another surface, taking in account the alpha channels
 *
 * @param origin The surface from which copy the piece
 * @param ox The x coordinate of the piece referred to the origin surface
 * @param oy The y coordinate of the piece referred to the origin surface
 * @param ow The width of the piece
 * @param oh The height of the piece
 * @param destination The surface to which copy the piece
 * @param dx The x coordinate where to copy the piece
 * @param dy The y coordinate where to copy the piece
 * @param do_copy If true, will just copy the origin surface without taking in account its alpha channel
 *
 * @return 0 on success, -1 on error
 */

int32_t blit_blend(TTOS_Surface *origin,int32_t ox, int32_t oy, int32_t ow, int32_t oh, TTOS_Surface *destination,
				  int32_t dx, int32_t dy, int32_t crop_x, int32_t crop_y, int32_t crop_w, int32_t crop_h, bool do_copy) {

	uint32_t x,y,o_jump,d_jump,oc,length;
	uint16_t *o_pointer,*d_pointer;
	uint8_t *oa_pointer,*da_pointer;
	uint16_t alpha1,alpha2,nc;

	if ((origin==NULL) || (destination==NULL)) {
		return -1;
	}

	/* Crop the piece to the original surface */
	ow = (ox+ow) > origin->w ? origin->w-ox : ow;
	oh = (oy+oh) > origin->h ? origin->h-oy : oh;

	if (((dx+ow)<=crop_x) || ((dy+oh)<=crop_y) || (dx>=(crop_x+crop_w)) || (dy>=(crop_y+crop_h))) {
		return 0; /* the piece falls out of the destination surface; nothing to paint */
	}

	/* Crop the piece to the destination surface */
	if (dx<crop_x) {
		ox-=(dx-crop_x);
		ow+=(dx-crop_x);
		dx=crop_x;
	}
	if (dy<crop_y) {
		oy-=(dy-crop_y);
		oh+=(dy-crop_y);
		dy=crop_y;
	}
	ow = (dx+ow) >= (crop_x+crop_w) ? crop_x+crop_w-dx : ow;
	oh = (dy+oh) >= (crop_y+crop_h) ? crop_y+crop_h-dy : oh;

	o_pointer=origin->pixels+(ox+oy*origin->w);
	d_pointer=destination->pixels+(dx+dy*destination->w);

	if (origin->alpha)
		oa_pointer=origin->alpha+(ox+oy*origin->w);
	else
		oa_pointer=NULL;

	if (destination->alpha)
		da_pointer=destination->alpha+(dx+dy*destination->w);
	else
		da_pointer=NULL;

	if ((!oa_pointer)||(do_copy)) {
		// If there's no alpha channel in the origin, use memcpy to do it faster

		o_jump=(origin->w);
		d_jump=(destination->w);
		length=2*ow;

		for(y=0;y<oh;y++) {
			memcpy(d_pointer,o_pointer,length);
			if (da_pointer) {
				memset(da_pointer,0,ow);
				da_pointer+=d_jump;
			}
			o_pointer+=o_jump;
			d_pointer+=d_jump;
		}
		return 0;
	}

	o_jump=(origin->w-ow);
	d_jump=(destination->w-ow);

	for(y=0;y<oh;y++) {
		for(x=0;x<ow;x++) {

			alpha1=(uint16_t)*(oa_pointer++);

			if (alpha1==255) {
				if (da_pointer)
					da_pointer++;
				o_pointer++;
				d_pointer++;
				continue;
			}

			nc=*(o_pointer++);

			/* fully opaque */
			if (!alpha1) {
				*(d_pointer++)=nc;
				if (da_pointer)
					*(da_pointer++)=0;
				continue;
			}

			/* semitransparent */

			if (da_pointer) {
				alpha2=(uint16_t)*da_pointer;
				alpha2*=alpha1;
				*(da_pointer++)=(uint8_t)((alpha2>>8)&0x00FF);
			}

			/*
			 * >>8 is an optimization, equivalent to divide with 256. Is used instead of the right value (255)
			 * because a shift is much faster that a division, and the result is nearly the same.
			 */

			oc=(uint32_t)(*(d_pointer));
			nc+=((((oc&0x0000F800)*((uint32_t)alpha1))>>8)&0x0000F800);
			nc+=((((oc&0x000007E0)*((uint32_t)alpha1))>>8)&0x000007E0);
			nc+=((((oc&0x0000001F)*((uint32_t)alpha1))>>8)&0x0000001F);
			*(d_pointer++)=nc;
		}
		o_pointer+=o_jump;
		d_pointer+=d_jump;
		oa_pointer+=o_jump;
		if (da_pointer)
			da_pointer+=d_jump;
	}

	return 0;
}


/**
 * Fills a surface with 0s, to put it black
 *
 * @param destination The surface to blank
 *
 * @return 0 on success, -1 on error
 *
 */

int32_t blit_erase_surface(TTOS_Surface *destination) {

	if (!destination) {
		return -1;
	}

	memset(destination->pixels,0,2*destination->w*destination->h);
	if (destination->alpha)
		memset(destination->alpha,255,destination->w*destination->h);
	return 0;

}


/**
 * Paints a pixel with the specified color and alpha
 *
 * @param destination The surface where to paint the pixel
 * @param x The x coordinate
 * @param y The y coordinate
 * @param color The desired color
 *
 * @return 0 on success, -1 on error
 */

int32_t blit_pixel(TTOS_Surface *destination,uint32_t x,uint32_t y, struct ttos_color *color) {

	uint32_t value;


	if ((!destination)||(!color)) {
		return -1;
	}

	value=x+y*destination->w;

	*(destination->pixels+value)=((((uint16_t)color->r)<<8)&0xF800)|((((uint16_t)color->g)<<3)&0x07E0)|((((uint16_t)color->b)>>3)&0x001F);
	if (destination->alpha) {
		*(destination->alpha+value)=255-color->a;
	}


	return 0;
}

/**
 * Paints four pixels with the specified color and alpha in (x,y), (w-x,y), (x,h-y) and (w-x,h-y)
 *
 * @param destination The surface where to paint the pixels
 * @param x The x coordinate
 * @param y The y coordinate
 * @param color The desired color (components must be premultiplied)
 *
 * @return 0 on success, -1 on error
 */

int32_t blit_multipixel(TTOS_Surface *destination,uint32_t x, uint32_t y, struct ttos_color *color) {

	uint32_t w,h;

	w=destination->w-1;
	h=destination->h-1;
	blit_pixel(destination,x,y,color);
	blit_pixel(destination,w-x,y,color);
	blit_pixel(destination,x,h-y,color);
	return(blit_pixel(destination,w-x,h-y,color));

}

/**
 * Fills a piece of a surface with the specified color
 *
 * @param destination The surface to paint
 * @param dx The x coordinate of the part to paint
 * @param dy The y coordinate of the part to paint
 * @param ow The width of the part to paint
 * @param oh The height of the part to paint
 * @param color The desired color
 * @param keep_bg If true, a blend with the current background will be made; if false, it will just fill the surface erasing
 * the background, but keeping the desired alpha channel (faster)
 *
 * @return 0 on success, -1 on error
 */

int32_t blit_fill(TTOS_Surface *destination,int32_t dx, int32_t dy, uint32_t ow, uint32_t oh, struct ttos_color *color, bool keep_bg) {

	uint32_t x,y,d_jump;
	uint16_t *d_pointer,oc,nc,alpha2;
	uint8_t *a_pointer,a;
	uint16_t color16,r,g,b;

	if ((!destination)||(!color)) {
		return -1;
	}

	if (!color->a) {
		return 0; // The color is fully transparent
	}

	if (((dx+ow)<=0) || ((dy+oh)<=0) || (dx>destination->w) || (dy>destination->h)) {
		return 0; // the piece falls out of the destination surface; nothing to paint
	}

	// Crop the piece to the destination surface
	if (dx<0) {
		ow+=dx;
		dx=0;
	}
	if (dy<0) {
		oh+=dy;
		dy=0;
	}
	ow = (dx+ow) > destination->w ? destination->w-dx : ow;
	oh = (dy+oh) > destination->h ? destination->h-dy : oh;

	d_jump=(destination->w-ow);

	r=(((uint16_t)color->r)*color->a)/255;
	g=(((uint16_t)color->g)*color->a)/255;
	b=(((uint16_t)color->b)*color->a)/255;
	a=255-color->a;
	color16=(((r)<<8)&0xF800)|(((g)<<3)&0x07E0)|(((b)>>3)&0x001F);
	d_pointer=destination->pixels+(dx+dy*destination->w);

	if ((!keep_bg)||(color->a==255)) {
		if (destination->alpha) {
			a_pointer=destination->alpha+(dx+dy*destination->w);
			for(y=0;y<oh;y++) {
				for(x=0;x<ow;x++) {
					*(d_pointer++)=color16;
					*(a_pointer++)=a;
				}
				d_pointer+=d_jump;
				a_pointer+=d_jump;
			}
		} else {
			for(y=0;y<oh;y++) {
				for(x=0;x<ow;x++) {
					*(d_pointer++)=color16;
				}
				d_pointer+=d_jump;
			}
		}
	} else {
		if (destination->alpha) {
			a_pointer=destination->alpha+(dx+dy*destination->w);
			for(y=0;y<oh;y++) {
				for(x=0;x<ow;x++) {
					alpha2=(uint16_t)*a_pointer;
					alpha2*=a;
					*(a_pointer++)=(uint8_t)((alpha2>>8)&0x00FF);
					/*
					 * >>8 is an optimization, equivalent to divide with 256. Is used instead of the right value (255)
					 * because a shift is much faster that a division, and the result is nearly the same.
					 */
					nc=color16;
					oc=(uint32_t)(*(d_pointer));
					nc+=((((oc&0x0000F800)*((uint32_t)a))>>8)&0x0000F800);
					nc+=((((oc&0x000007E0)*((uint32_t)a))>>8)&0x000007E0);
					nc+=((((oc&0x0000001F)*((uint32_t)a))>>8)&0x0000001F);
					*(d_pointer++)=nc;
				}
				d_pointer+=d_jump;
				a_pointer+=d_jump;
			}
		} else {
			for(y=0;y<oh;y++) {
				for(x=0;x<ow;x++) {
					/*
					 * >>8 is an optimization, equivalent to divide with 256. Is used instead of the right value (255)
					 * because a shift is much faster that a division, and the result is nearly the same.
					 */
					nc=color16;
					oc=(uint32_t)(*(d_pointer));
					nc+=((((oc&0x0000F800)*((uint32_t)a))>>8)&0x0000F800);
					nc+=((((oc&0x000007E0)*((uint32_t)a))>>8)&0x000007E0);
					nc+=((((oc&0x0000001F)*((uint32_t)a))>>8)&0x0000001F);
					*(d_pointer++)=nc;
				}
				d_pointer+=d_jump;
			}
		}
	}

	return 0;
}
