/*
 * Copyright 2010 (C) Raster Software Vigo (Sergio Costas)
 * This file is part of MINIBUS
 *
 * MINIBUS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * MINIBUS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/*
 * list.c
 *
 *  Created on: 21/08/2010
 *      Author: raster
 */

#include <stdio.h>
#include <stdlib.h>

#include "minibus.h"

/**
 * Initialize a double-linked list
 *
 * @param list The list to initialize
 * @param destroy_callback A pointer to a function which frees the memory used by an element contained in the list.
 * Called each time an element is removed from the list. If NULL is passed, no function will be called.
 */

void minibus_list_init(struct minibus_list *list, void (*destroy_callback) (void *element), bool (*sort_callback) (void *element1, void *element2)) {

	list->list=NULL;
	list->last=NULL;
	list->destroy_callback=destroy_callback;
	list->sort_callback=sort_callback;
}

/**
 * Finds the container for a particular data type in a list
 *
 * @param list The list into which search the data
 * @param data The data to find
 * @return A minibus_list_element containing that data, or NULL if not found
 */

struct minibus_list_element *minibus_find_data(struct minibus_list *list, void *data) {

	struct minibus_list_element *element;

	for(element=list->list;element;element=element->next) {
		if (data==element->data) {
			return (element);
		}
	}
	return NULL;

}

/**
 * Adds a node to a double-linked list
 *
 * @param list The list where the node will be added
 * @param data The data to insert in the node
 *
 * @return The struct element, or NULL on failure
 */

struct minibus_list_element * minibus_list_add(struct minibus_list *list,void *data) {

	struct minibus_list_element *element;

	if (!list) {
		return NULL;
	}

	element=(struct minibus_list_element *)malloc(sizeof(struct minibus_list_element));
	if (!element)
		return NULL;

	if (list->last) {
		list->last->next=element;
	} else {
		list->list=element;
	}
	element->data=data;
	element->next=NULL;
	element->prev=list->last;
	element->list=list;
	element->refcount=1;
	list->last=element;
	return element;

}


/**
 * Appends all the elements from a list into another, and empties the first
 *
 * @param destination List into which the elements will be appended
 * @param origin List from which the elements will be taken
 * @return 0 if everything went fine; -2 if the lists are incompatible, and -1 if the lists are incorrect
 */

int32_t minibus_list_append(struct minibus_list *destination, struct minibus_list *origin) {

	struct minibus_list_element *element;

	if ((!origin)||(!destination)) {
		return -1;
	}

	if (origin->destroy_callback!=destination->destroy_callback) {
		return -2;
	}

	if (!origin->list) {
		return(0);
	}

	if (!destination->last) {
		destination->last=origin->last;
	} else {
		destination->list->prev=origin->last;
	}
	origin->last->next=destination->list;
	destination->list=origin->list;
	origin->list=NULL;
	origin->last=NULL;

	for(element=destination->list;element;element=element->next) {
		element->list=destination;
	}

	return 0;
}


/**
 * Adds an element to a double-linked list keeping it in increasing order
 *
 * @param list The list where the node will be added
 * @param data The data to insert in the node
 * @param order The place into which insert the node
 *
 * @return The container element on success, or NULL on failure
 */

struct minibus_list_element * minibus_list_add_sorted(struct minibus_list *list,void *data) {

	struct minibus_list_element *element,*loop;

	if (!list) {
		return NULL;
	}

	if (!list->sort_callback) {
		return NULL;
	}

	element=(struct minibus_list_element *)malloc(sizeof(struct minibus_list_element));
	if (!element) {
		return NULL;
	}

	element->data=data;
	element->list=list;
	element->refcount=1;

	if (!list->list) {
		list->last=element;
		list->list=element;
		element->prev=NULL;
		element->next=NULL;
		return element;
	}

	if (list->sort_callback(data,list->list->data)) {
		element->prev=NULL;
		element->next=list->list;
		list->list->prev=element;
		list->list=element;
		return element;
	}

	for(loop=list->list;loop->next!=NULL;loop=loop->next) {
		if (list->sort_callback(data,loop->next->data)) {
			element->next=loop->next;
			element->prev=loop;
			loop->next=element;
			element->next->prev=element;
			return element;
		}
	}
	element->next=NULL;
	element->prev=list->last;
	element->prev->next=element;
	list->last=element;
	return element;
}


/**
 * Increments the reference counter for a node.
 *
 * @param object The list node to which increment the reference counter
 *
 * @return The new value, or -1 if there's an error
 */

uint32_t minibus_list_incref(struct minibus_list_element *object) {

	if (object==NULL) {
		return 0;
	}

	object->refcount++;
	return (object->refcount);
}

/**
 * Decrements the reference counter for a node, and, if the counter reaches 0, frees the node and the data (if desired).
 *
 * @param object The list node to which increment the reference counter
 *
 * @return The new reference count value, or 0 if there's an error or the node is liberated
 */

uint32_t minibus_list_decref(struct minibus_list_element *object) {

	uint32_t value;

	if (object==NULL) {
		return 0;
	}

	object->refcount--;
	value=object->refcount;
	if (!value) {
		minibus_list_remove(object,true);
	}
	return (value);
}


/**
 * Removes an element from a list, and call, if desired, a callback for free the data
 *
 * @param element The element to remove from the list
 */

void minibus_list_remove(struct minibus_list_element *element, bool free_data) {

	void (* destroy_callback)(void *);

	destroy_callback=element->list->destroy_callback;

	if (element->next) {
		element->next->prev=element->prev;
	} else {
		element->list->last=element->prev;
	}

	if (element->prev) {
		element->prev->next=element->next;
	} else {
		element->list->list=element->next;
	}

	if ((destroy_callback)&&(free_data)) {
		destroy_callback(element->data);
	}

	free(element);

}


/**
 * Removes an element from a list based on its pointer to data, and call, if desired, a callback for free the data
 *
 * @param list The list containing the data
 * @param data The data to remove from the list
 *
 * @return 0 if success; -1 if the data element is not in the list
 */

int32_t minibus_list_remove_data(struct minibus_list *list, void *data, bool free_data) {

	struct minibus_list_element *element;

	for(element=list->list;element!=NULL;element=element->next) {
		if (element->data==data) {
			minibus_list_remove(element,free_data);
			return 0;
		}
	}
	return -1;
}


/**
 * Removes all elements from a list, calling, if desired, a callback to free the data in each node
 *
 * @param list The list to empty
 */

void minibus_list_free(struct minibus_list *list) {

	void (* destroy_callback)(void *);
	struct minibus_list_element *element,*next;

	if (!list) {
		return;
	}

	destroy_callback=list->destroy_callback;

	for(element=list->list;element!=NULL;element=next) {
		next=element->next;
		if (destroy_callback) {
			destroy_callback(element->data);
		}
		free(element);
	}
	list->list=NULL;
	list->last=NULL;
}

