/*
 * Copyright 2010 (C) Raster Software Vigo (Sergio Costas)
 * This file is part of MINIBUS
 *
 * MINIBUS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * MINIBUS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/*
 * rw.c
 *
 *  Created on: 21/08/2010
 *      Author: raster
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "minibus.h"

/**
 * Stores a byte (8 bits) in the specified connection buffer and increments the pointers to store the next data.
 *
 * @param object Connection object where to store the data
 * @param value The value to be stored
 *
 * @return 0 on success; -1 if object is invalid; -2 if the queue is full
 */

int32_t minibus_write_byte(struct minibus_connection *object, uint8_t value) {

	uint8_t *pointer;

	if (object==NULL) {
		return -1;
	}

	if (object->s_size>65534) {
		return -2;
	}

	pointer=(object->send_message)+(object->s_pointer);
	object->s_pointer+=2;
	object->s_size+=2;

	*(pointer++) = MINIBUS_TYPE_INT8; // byte
	*(pointer) = value;
	return 0;
}


/**
 * Stores a short (16 bits) in the specified connection buffer and increments the pointers to store the next data.
 *
 * @param object Connection object where to store the data
 * @param value The value to be stored
 *
 * @return 0 on success; -1 if object is invalid; -2 if the queue is full
 */

int32_t minibus_write_short(struct minibus_connection *object, uint16_t value) {

	uint8_t *pointer;

	if (object==NULL) {
		return -1;
	}

	if (object->s_size>65533) {
		return -2;
	}

	pointer=(object->send_message)+(object->s_pointer);
	object->s_pointer+=3;
	object->s_size+=3;

	*(pointer++) = MINIBUS_TYPE_INT16; // short
	*(pointer++) = (uint8_t) ((value>>8)&0x0FF);
	*(pointer)   = (uint8_t) (value&0x0FF);
	return 0;
}


/**
 * Stores a long (32 bits) in the specified connection buffer and increments the pointers to store the next data.
 *
 * @param object Connection object where to store the data
 * @param value The value to be stored
 *
 * @return 0 on success; -1 if object is invalid; -2 if the queue is full
 */

int32_t minibus_write_long(struct minibus_connection *object, uint32_t value) {

	uint8_t *pointer;

	if (object==NULL) {
		return -1;
	}

	if (object->s_size>65531) {
		return -2;
	}

	pointer=object->send_message+object->s_pointer;
	object->s_pointer+=5;
	object->s_size+=5;

	*(pointer++) = MINIBUS_TYPE_INT32; // long
	*(pointer++) = (uint8_t) ((value>>24)&0x0FF);
	*(pointer++) = (uint8_t) ((value>>16)&0x0FF);
	*(pointer++) = (uint8_t) ((value>>8)&0x0FF);
	*(pointer)   = (uint8_t) (value&0x0FF);
	return 0;
}


/**
 * Stores a double long (64 bits) in the specified connection buffer and increments the pointers to store the next data.
 *
 * @param object Connection object where to store the data
 * @param value The value to be stored
 *
 * @return 0 on success; -1 if object is invalid; -2 if the queue is full
 */

int32_t minibus_write_double(struct minibus_connection *object, uint64_t value) {

	uint8_t *pointer;

	if (object==NULL) {
		return -1;
	}

	if (object->s_size>65527) {
		return -2;
	}

	pointer=object->send_message+object->s_pointer;
	object->s_pointer+=9;
	object->s_size+=9;

	*(pointer++) = MINIBUS_TYPE_INT32; // long
	*(pointer++) = (uint8_t) ((value>>56)&0x0FF);
	*(pointer++) = (uint8_t) ((value>>48)&0x0FF);
	*(pointer++) = (uint8_t) ((value>>40)&0x0FF);
	*(pointer++) = (uint8_t) ((value>>32)&0x0FF);
	*(pointer++) = (uint8_t) ((value>>24)&0x0FF);
	*(pointer++) = (uint8_t) ((value>>16)&0x0FF);
	*(pointer++) = (uint8_t) ((value>>8)&0x0FF);
	*(pointer)   = (uint8_t) (value&0x0FF);
	return 0;
}


/**
 * Stores a string (ended with a 0, ASCIIZ) in the specified connection buffer and increments the pointers to store
 * the next data.
 *
 * @param object Connection object where to store the data
 * @param string The string to be stored
 *
 * @return 0 on success; -1 if object is invalid; -2 if the queue is full
 */

int32_t minibus_write_string(struct minibus_connection *object, Cstring string) {

	uint32_t size;

	if (object==NULL) {
		return -1;
	}

	if (!string) {
		return -1;
	}

	size=1+strlen((char *)string);
	if ((object->s_size) > (65535-size-1)) {
		return -2;
	}

	*(object->send_message+object->s_pointer) = MINIBUS_TYPE_STRING; //string
	memcpy(object->send_message+object->s_pointer+1,string,size);
	object->s_pointer+=size+1;
	object->s_size+=size+1;
	return 0;
}


/**
 * Stores a BLOB (binary object) in the specified connection buffer and increments the pointers to store
 * the next data.
 *
 * @param object Connection object where to store the data
 * @param data The sequence of bytes to be stored
 * @param size The size of the BLOB
 *
 * @return 0 on success; -1 if object is invalid; -2 if the queue is full
 */

int32_t minibus_write_blob(struct minibus_connection *object, uint8_t * data, uint32_t size) {

	uint8_t *pointer;

	if (object==NULL) {
		return -1;
	}

	if ((object->s_size) > (65535-size-3)) {
		return -2;
	}

	pointer=object->send_message+object->s_pointer;
	*(pointer++) = MINIBUS_TYPE_BLOB; // BLOB
	*(pointer++) = (uint8_t) ((size>>8)&0x0FF);
	*(pointer++) = (uint8_t) (size&0x0FF);
	memcpy(pointer, data, size);
	object->s_pointer+=size+3;
	object->s_size+=size+3;
	return 0;
}


/**
 * Reads a value from the specified connection buffer and increments the pointers to be ready for read the
 * next value in a received message.
 *
 * @param object Connection object from where to read the byte
 * @param data Object where the data will be stored
 *
 * @return 0 on success; -1 if object or data are invalid; -2 if there's no more data available
 */

int32_t minibus_read_data(struct minibus_connection *object,struct minibus_data_element *data) {

	uint8_t type;
	uint8_t *pointer;
	int size;

	if (object==NULL) {
		return -1;
	}

	if (data==NULL) {
		return -1;
	}

	if (!object->received_message) {
		return -2;
	}

	if (object->r_pointer>=object->r_size+2) {
		data->type=MINIBUS_TYPE_EMPTY;
		return -2;
	}

	pointer=object->received_message+object->r_pointer;
	type=*(pointer++);
	data->type=type;
	switch(type) {
	case MINIBUS_TYPE_INT8:
		data->value8=*(pointer);
		object->r_pointer+=2;
		break;
	case MINIBUS_TYPE_INT16:
		data->value16=((((uint16_t)(*pointer))<<8)&0xFF00)|(((uint16_t)*(pointer+1))&0x00FF);
		object->r_pointer+=3;
		break;
	case MINIBUS_TYPE_INT32:
		data->value32=((((uint32_t) *(pointer))  <<24)&0xFF000000)+
					  ((((uint32_t) *(pointer+1))<<16)&0x00FF0000)+
					  ((((uint32_t) *(pointer+2))<<8) &0x0000FF00)+
					   (((uint32_t) *(pointer+3))     &0x000000FF);
		object->r_pointer+=5;
		break;
	case MINIBUS_TYPE_STRING:
		size=strlen((char *)pointer);
		data->string=(Cstring)pointer;
		data->stringsize=size;
		object->r_pointer+=2+size;
		break;
	case MINIBUS_TYPE_BLOB:
		size=((((uint16_t)(*pointer))<<8)&0xFF00)|(((uint16_t)*(pointer+1))&0x00FF);
		data->blob=pointer+2;
		data->blobsize=size;
		object->r_pointer+=3+size;
		break;
	default:
		data->type=MINIBUS_TYPE_ERROR;
	}

	return(0);
}

/**
 * Returns a sequence of parameters received from a connection, in the specified format.
 *
 * @param connection The connection from which the data will be read
 * @param format A string with the expected type for each parameter: b=byte; s=short; l=long; c=string; n=blob; a=any format;
 * @param error A pointer where the function will store the cause of an error
 *
 * @return An array of elements (ended with a MINIBUS_TYPE_EMTPY entry), or NULL on error
 *
 */

struct minibus_data_element *minibus_read_params(struct minibus_connection *connection, char *format,int32_t *error) {

	struct minibus_data_element *object,*tmp;
	int32_t size;
	int8_t *pointer;

	size=strlen(format);
	object=malloc((size+1)*sizeof(struct minibus_data_element));
	if (!object) {
		if (error)
			*error=-1; // can't reserve memory
		return NULL;
	}

	for(pointer=(int8_t *)format,tmp=object;*pointer!=0;pointer++) {
		if (0!=minibus_read_data(connection,tmp)) {
			if (error)
				*error=-2; // error while reading a parameter
			free(object);
			return NULL;
		}
		switch(*pointer) {
		case 'b':
			if (tmp->type!=MINIBUS_TYPE_INT8) {
				if (error)
					*error=-3; // incorrect type
				free(object);
				return (NULL);
			}
		break;
		case 's':
			if (tmp->type!=MINIBUS_TYPE_INT16) {
				if (error)
					*error=-3; // incorrect type
				free(object);
				return (NULL);
			}
		break;
		case 'l':
			if (tmp->type!=MINIBUS_TYPE_INT32) {
				if (error)
					*error=-3; // incorrect type
				free(object);
				return (NULL);
			}
		break;
		case 'c':
			if (tmp->type!=MINIBUS_TYPE_STRING) {
				if (error)
					*error=-3; // incorrect type
				free(object);
				return (NULL);
			}
		break;
		case 'n':
			if (tmp->type!=MINIBUS_TYPE_BLOB) {
				if (error)
					*error=-3; // incorrect type
				free(object);
				return (NULL);
			}
		break;
		case 'a': // don't care
		break;
		default:
			if (error)
				*error=-4; // undefined type
			free(object);
			return(NULL);
		break;
		}
		tmp++;
	}
	tmp->type=MINIBUS_TYPE_EMPTY;
	return(object);
}


/**
 * Waits for a message in an specified port.
 *
  * @param conn The connection into which wait
 * @param port Port to wait
 * @param do_callback If true, the callbacks for other ports will be called; if false, the messages will be stored and called
 * as soon as the main program enter/returns to a main loop
 *
 * @return 0: message fully read; 1: timeout; -1: the object is invalid; -2: there was an error while receiving
 *
 */

int32_t minibus_wait_for_port(struct minibus_connection *conn, uint32_t port, bool do_callback) {


	int serversock;
	int32_t retval;
	struct minibus_wait_data port_conn;
	struct minibus_master *object;
	struct minibus_message_element *message;
	bool old_loop;

	if(!conn) {
		return -1;
	}

	if (port==0x00) {
		return -2; // Main port won't be waitable
	}

	object=conn->master_queue;

	if(!object) {
		return -1;
	}

	if (object->server) {
		serversock=object->id_socket;
	} else {
		serversock=0;
	}

	if (!object->polling) {
		object->polling=minibus_fill_polling(object,serversock);
		if (object->polling==NULL) {
			return -1;
		}
	}

	old_loop=object->loop;
	object->loop=true;
	while(object->loop) {
		retval=minibus_read_message(object,serversock,&port_conn,do_callback);
		if (retval<0) {
			break;
		}
		if (retval==1) { // a message arrived; do its callback
			if ((port_conn.port==port)&&(port_conn.conn==conn)) {
				retval=0;
				break;
			}
			if ((do_callback)||(port_conn.port==0)) { // Port 0 will always be called
				minibus_do_callback(object,&port_conn);
			} else {
				message=minibus_create_receive_message_element(port_conn.conn,port_conn.port);
				if (message) {
					minibus_list_add(&conn->receive_buffer,message);
				}
			}
		}
		retval=0;
	}
	if ((!object->loop) || (!old_loop)) {
		object->loop=false;
	}
	return (retval);
}


