/*
 * Copyright 2011 Raster Software Vigo
 *
 * This file is part of ttos
 *
 * ttos is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * ttos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * sharemem.c
 *
 *  Created on: 20/02/2011
 *      Author: raster
 */

#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>

#include "sharemem.h"
#include "uid.h"
#include "global_vars.h"
#include "widgets.h"
#include "other.h"

struct _shared_mem* sharemem_new(Cstring file, uint16_t width, uint16_t height, bool has_alpha) {

	struct _shared_mem *object;
	int f;

	object=malloc(sizeof(struct _shared_mem));
	if (!object) {
		return NULL;
	}

	object->has_alpha=has_alpha;
	object->width=width;
	object->height=height;
	object->uid=uid_get_uid(ttos_uids, TTOS_SHMEM_UID_MASK);
	if (!object->uid) {
		free(object);
		return NULL;
	}
	printf("Nuevo sharemem %X\n",object->uid);
#ifdef USE_ARM
	f=open(file,O_RDWR, S_IRWXU | S_IRWXG);
#else
	f=shm_open(file,O_RDWR, S_IRWXU | S_IRWXG);
#endif
	if (f<0) {
		free(object);
		return NULL;
	}


	if (has_alpha) {
		object->final_size=3*width*height;
	} else {
		object->final_size=2*width*height;
	}

	object->pixels=mmap(NULL,object->final_size,PROT_READ | PROT_WRITE, MAP_SHARED, f,0);
	close(f);
	shm_unlink(file);
	if (!object->pixels) {
		free(object);
		return NULL;
	}
	return (object);
}

void sharemem_free(void *element) {

	struct _shared_mem *object;

	if (!element) {
		return;
	}

	object=element;

	uid_release_uid(ttos_uids,object->uid);
	munmap(object->pixels,object->final_size);
	free(object);

}


/**
 * Creates a new sharemem object. It receives a message with the following fields:
 *
 * int16 width The width for the associated drawable
 * int16 height The height for the associated drawable
 * int8 has_alpha If the associated drawable will have alpha channel
 * string The file name associated with the drawable
 */

uint32_t sharemem_create(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	uint16_t w,h;
	int32_t error;
	Cstring filename;
	bool has_alpha;
	struct _shared_mem *object;
	struct minibus_data_element *data;
	struct _client_data *client;

	client=connection->data;

	data=minibus_read_params(connection,"ssbc",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	w=data[0].value16;
	h=data[1].value16;
	has_alpha=data[2].value8 ? true : false;
	filename=strdup(data[3].string);
	free(data);

	object=sharemem_new(filename,w,h,has_alpha);
	free(filename);
	if (!object) {
		other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
		return -1;
	}

	minibus_list_add(&client->sharemem,object);
	other_send_uid(connection,answer_port,object->uid);
	return 0;
}

/**
 * Destroys a sharemem object. It receives a message with the following fields:
 * int32 UID The sharemem UID to destroy
 */

uint32_t sharemem_destroy(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	uint32_t uid;
	int32_t error;
	struct _shared_mem *object;
	struct minibus_data_element *data;
	struct _client_data *client;
	struct minibus_list_element *element,*element2;
	struct _drawable_widget *drawable;
	struct _widget *widget;
	struct _tab *tab;

	client=connection->data;

	data=minibus_read_params(connection,"l",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	uid=data[0].value32;
	free(data);

	for(element=client->sharemem.list;element;element=element->next) {
		object=element->data;
		if (object->uid==uid) {
			break;
		}
		object=NULL;
	}
	if (!object) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		return -1;
	}

	for(element=client->tabs.list;element;element=element->next) {
		tab=element->data;
		for(element2=tab->widgets.list;element2;element2=element2->next) {
			widget=element2->data;
			if (widget->type!=TTOS_WIDGET_DRAWABLE) {
				continue;
			}
			drawable=element2->data;
			if (drawable->bitmap->pixels==object->pixels) {
				other_send_error(connection,answer_port, TTOS_ERROR_SHMEM_IN_USE);
				return -1;
			}
		}
	}

	// The shared memory object is not in use

	minibus_list_remove_data(&client->sharemem,object,true);
	other_send_error(connection,answer_port, TTOS_ERROR_NO_ERROR);
	return 0;
}
