/*
 * Copyright 2011 Raster Software Vigo
 *
 * This file is part of ttos
 *
 * ttos is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * ttos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * tabs.c
 *
 *  Created on: 21/09/2010
 *      Author: raster
 */

#include <stdlib.h>
#include <string.h>

#include "ttos_lib/ttos_structs.h"
#include "other.h"
#include "tabs.h"
#include "widgets.h"
#include "timer.h"

/**
 * Creates a new tab in the client
 */

uint32_t tabs_create_tab(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	struct _tab *object;
	struct _client_data *client;
	uint32_t tab_uid;

	tab_uid=uid_get_uid(ttos_uids,TTOS_TAB_UID_MASK);
	if (tab_uid==0) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		return (-1);
	}

	client=(struct _client_data *)minibus_get_connection_data(connection);

	object=malloc(sizeof(struct _tab));
	if (!object) {
		other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
		return (-1);
	}

	object->client=client;
	object->tab_id=(uint16_t)(tab_uid&0x0000FFFF);
	object->swap_tab=NULL;
	object->to_delete=false;
	minibus_list_init(&object->widgets,widgets_free_widget,NULL);
	object->widget_id_counter=0;
	object->color.b=0;
	object->color.g=0;
	object->color.r=0;
	object->color.a=0; // by default, transparent
	object->x_offset=0;
	object->y_offset=0;
	object->minx=0;
	object->miny=0;
	object->maxx=windows[client->window].w-1;
	object->maxy=windows[client->window].h-1;
	object->margin_up=0;
	object->margin_down=0;
	object->margin_left=0;
	object->margin_right=0;
	object->xswap=0;
	object->yswap=0;
	object->sensitive=true;
	object->xtmp=0;
	object->ytmp=0;
	object->valid_tmp=false;
	object->client->edit_tab=object;
	minibus_list_add(&client->tabs,object);
	other_send_uid(connection,answer_port,tab_uid);
	return (0);
}


/**
 * Returns the tab being showed now
 *
 */

uint32_t tabs_get_current_tab(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	struct _client_data *client;

	client=(struct _client_data *)minibus_get_connection_data(connection);
	if (!client->show_tab) {
		other_send_error(connection,answer_port,TTOS_ERROR_NO_VISIBLE_TAB);
		return -1;
	}
	other_send_uid(connection,answer_port,TTOS_TAB_UID_MASK|((uint32_t)client->show_tab->tab_id));
	return (0);
}



/**
 * Forces a specific virtual size for a tab
 *
 * Receives a message with five uint16_t and one uint8_t, containing the uid to assign to the new tab, the x,y coordinates, and the width
 * and height, and finally a boolean which specifies if that value must be forced (true), or will be only minimum values
 */

uint32_t tabs_set_size(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	struct minibus_data_element *data;
	struct minibus_list_element *element;
	struct _widget *widget;
	struct _client_data *client;
	int32_t error;
	uint16_t tab_id;
	int16_t x,y;
	uint16_t w,h;
	uint8_t force;

	data=minibus_read_params(connection,"sssssb",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	tab_id=data[0].value16;
	x=data[1].value16;
	y=data[2].value16;
	w=data[3].value16;
	h=data[4].value16;
	force=data[5].value8;
	free(data);

	if (false==tabs_select_edit_tab(connection,tab_id)) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		return -1;
	}

	client=(struct _client_data *)minibus_get_connection_data(connection);

	if(force) {
		client->edit_tab->minx=x;
		client->edit_tab->miny=y;
		client->edit_tab->maxx=x+w-1;
		client->edit_tab->maxy=y+h-1;
	} else {

		// Recalculate the minimun and maximum values

		client->edit_tab->minx=x;
		client->edit_tab->miny=y;
		client->edit_tab->maxx=x+w-1;
		client->edit_tab->miny=y+h-1;

		for(element=client->edit_tab->widgets.list;element;element=element->next) {
			widget=element->data;
			tabs_set_minmax(client->edit_tab,widget);
		}
	}
	other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
	return 0;
}


/**
 * Sets the minimum and maximum X and Y values for a tab. Called each time a widget is added
 *
 * @param tab The tab to adjust
 * @param x The X coordinate of the widget being added
 * @param y The Y coordinate of the widget being added
 * @param w The width of the widget being added
 * @param h The height of the widget being added
 */

void tabs_set_minmax(struct _tab *tab,struct _widget *widget) {

	int16_t x,y,w,h;

	x=widget->ox-tab->margin_left;
	y=widget->oy-tab->margin_up;
	w=widget->w+tab->margin_right;
	h=widget->h+tab->margin_down;

	if (!widget->h_fixed) {
		if (x<tab->minx) {
			tab->minx=x;
		}
		if ((x+w-1)>tab->maxx) {
			tab->maxx=x+w-1;
		}
	}

	if (!widget->v_fixed) {
		if (y<tab->miny) {
			tab->miny=y;
		}
		if ((y+h-1)>tab->maxy) {
			tab->maxy=y+h-1;
		}
	}
}


/**
 * Sets the margins for widgets in an specified tab
 *
 */


uint32_t tabs_set_margins(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	struct minibus_data_element *data;
	struct _client_data *client;
	int32_t error;
	uint16_t tab_id;
	uint16_t mu,md,ml,mr;

	data=minibus_read_params(connection,"sssss",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	tab_id=data[0].value16;
	mu=data[1].value16;
	md=data[2].value16;
	ml=data[3].value16;
	mr=data[4].value16;
	free(data);

	if (false==tabs_select_edit_tab(connection,tab_id)) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		return -1;
	}

	client=(struct _client_data *)minibus_get_connection_data(connection);

	client->edit_tab->margin_up=mu;
	client->edit_tab->margin_down=md;
	client->edit_tab->margin_left=ml;
	client->edit_tab->margin_right=mr;
	other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
	return 0;


}


/**
 * Sets a tab's background color
 *
 */


uint32_t tabs_set_bg_color(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	struct minibus_data_element *data;
	struct _client_data *client;
	int32_t error;
	uint16_t tab_id;
	uint8_t r,g,b,a;

	data=minibus_read_params(connection,"sbbbb",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	tab_id=data[0].value16;
	r=data[1].value8;
	g=data[2].value8;
	b=data[3].value8;
	a=data[4].value8;
	free(data);

	if (false==tabs_select_edit_tab(connection,tab_id)) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		return -1;
	}

	client=(struct _client_data *)minibus_get_connection_data(connection);

	client->edit_tab->color.r=r;
	client->edit_tab->color.g=g;
	client->edit_tab->color.b=b;
	client->edit_tab->color.a=a;
	other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
	return 0;
}



/**
 * Destroys a tab in the client
 *
 * Receives a message with a uint16_t, containing the uid of the tab to destroy, and a uint32_t with flags (TTOS_TAB_DELAY_DESTROY)
 */


uint32_t tabs_destroy_tab(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	struct _client_data *client;
	struct minibus_data_element *data;
	struct minibus_list_element *element,*element_next;
	struct _tab *object;
	uint16_t tab_id;
	int32_t error;
	uint32_t flags;

	data=minibus_read_params(connection,"sl",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	tab_id=data[0].value16;
	flags=data[1].value32;
	free(data);

	client=(struct _client_data *)minibus_get_connection_data(connection);

	for(element=client->tabs.list;element!=NULL;element=element_next) {
		element_next=element->next;
		object=element->data;
		if (object->tab_id==tab_id) {
			if (client->show_tab==object) {
				if (flags&TTOS_TAB_DELAY_DESTROY) {
					object->to_delete=true;
					other_send_error(connection,answer_port,TTOS_ERROR_DESTROY_DELAYED);
					return 0;
				} else {
					client->show_tab=NULL;
				}
			}
			if (client->edit_tab==object) {
				client->edit_tab=NULL;
			}
			minibus_list_remove(element,true);
			other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
			uid_release_uid(ttos_uids,(uint32_t)(tab_id));
			return 0;
		}
	}
	other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);

	return 0;
}


/**
 * Destroys a tab structure, freeing all its elements
 *
 * @param element The tab to destroy
 *
 */

void tabs_free_tab(void *object) {

	struct _tab *tab=object;
	struct minibus_list_element *element,*element_next;
	struct _tab_scroll *scroll;

	if (!tab) {
		return;
	}

	if (!object) {
		return;
	}

	for(element=tab_scroll.list;element;element=element_next) {
		element_next=element->next;
		scroll=element->data;
		if (!scroll->tab) {
			continue;
		}
		if (scroll->tab->tab_id==tab->tab_id) { // This tab is being scrolled, so we must remove it
			minibus_list_remove(element,true);
		}
		if (scroll->tab->swap_tab==object) { // there's a tab scrolling over this tab, so we must null it
			scroll->tab->swap_tab=NULL;
		}
	}
	minibus_list_free(&tab->widgets);

	free(tab);
}


/**
 * Sets the specified tab as the default editing tab for the specified connection
 *
 * @param connection The connection which contains the tab
 * @param tab_id The tab's identifier
 *
 * @return true if all went fine; false if there was an error
 */

bool tabs_select_edit_tab(struct minibus_connection *connection, uint16_t tab_id) {

	struct _client_data *client;
	struct minibus_list_element *element;
	struct _tab *tab;

	if (!connection) {
		return false;
	}

	client=connection->data;
	if ((client->edit_tab)&&(client->edit_tab->tab_id==tab_id)) { // The current edit tab is already the desired tab
		return true;
	}

	for(element=client->tabs.list;element!=NULL;element=element->next) {
		tab=element->data;
		if (tab->tab_id==tab_id) {
			client->edit_tab=tab;
			return true;
		}
	}
	return false;
}


/**
 * Sets the specified tab as the current tab for the specified connection
 *
 * @param connection The connection which contains the tab
 * @param tab_id The tab's identifier
 *
 * @return true if all went fine; false if there was an error
 */

uint32_t tabs_select_show_tab(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	struct _client_data *client;
	struct minibus_list_element *element;
	struct _tab *tab,*old_tab;
	struct minibus_data_element *data;
	uint16_t tab_id;
	uint8_t swap_mode;
	uint32_t event_port;
	int32_t error;
	struct _tab_scroll *swap_tab;


	if (!connection) {
		return -1;
	}

	data=minibus_read_params(connection,"sbl",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	tab_id=data[0].value16;
	swap_mode=data[1].value8;
	event_port=data[2].value32;
	free(data);

	client=connection->data;
	if ((client->show_tab)&&(client->show_tab->tab_id==tab_id)) { // The current show tab is already the desired tab
		other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
		return 0;
	}

	for(element=client->tabs.list;element!=NULL;element=element->next) {
		tab=element->data;
		if (tab->tab_id==tab_id) {
			break;
		}
	}

	if (!element) {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		return -1;
	}

	old_tab=client->show_tab;
	client->show_tab=tab;
	tab->valid_tmp=false;
	switch(swap_mode) {
	case TTOS_TAB_SWAP_MODE_REPLACE:
		other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
		if (event_port!=TTOS_NO_ANSWER) {
			minibus_start_message(write_conn,event_port,TTOS_NO_ANSWER);
			minibus_write_long(write_conn,TTOS_SR_TAB_SWAP_END);
			minibus_write_short(write_conn,tab_id);
			minibus_server_send(write_conn,connection,false);
		}
		if (old_tab && (old_tab->to_delete)) {
			uid_release_uid(ttos_uids,(uint32_t)old_tab->tab_id);
			minibus_list_remove_data(&client->tabs,old_tab,true);
		}
		return 0;
	break;
	case TTOS_TAB_SWAP_MODE_SCROLL_RIGHT:
	case TTOS_TAB_SWAP_MODE_SCROLL_LEFT:
		swap_tab=tabs_init_tab_scroll(client->show_tab,old_tab,0,0,event_port,swap_mode,true);
		if(!swap_tab) {
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		}
		client->show_tab->swap_tab=old_tab;
		if (old_tab) {
			old_tab->swap_tab=NULL;
			old_tab->xswap=0;
			old_tab->yswap=0;
		}
		if (swap_mode==TTOS_TAB_SWAP_MODE_SCROLL_LEFT) {
			tab->xswap=windows[tab->client->window].w;
		} else {
			tab->xswap=-windows[tab->client->window].w;
		}
		minibus_list_add(&tab_scroll,swap_tab);
		timer_start_scroll();
		other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
		return 0;
	break;
	case TTOS_TAB_SWAP_MODE_SCROLL_UP:
	case TTOS_TAB_SWAP_MODE_SCROLL_DOWN:
		swap_tab=tabs_init_tab_scroll(client->show_tab,old_tab,0,0,event_port,swap_mode,true);
		if(!swap_tab) {
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		}
		client->show_tab->swap_tab=old_tab;
		if (old_tab) {
			old_tab->swap_tab=NULL;
			old_tab->xswap=0;
			old_tab->yswap=0;
		}
		if (swap_mode==TTOS_TAB_SWAP_MODE_SCROLL_UP) {
			tab->yswap=windows[tab->client->window].h;
		} else {
			tab->yswap=-windows[tab->client->window].h;
		}

		minibus_list_add(&tab_scroll,swap_tab);
		timer_start_scroll();
		other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
		return 0;
	break;
	}
	other_send_error(connection,answer_port,TTOS_ERROR_INVALID_SWAP_MODE);
	return 0;
}


/**
 * Repaints a tab
 */


void tabs_paint_tab(struct _tab *tab) { //, int16_t x, int16_t y, uint16_t w, uint16_t h) {

	struct minibus_list_element *element;
	struct _widget *widget;
	int16_t final_x,final_y;
	struct _window *c_window,tmp_window;
	int16_t x;
	int16_t y;
	uint16_t w;
	uint16_t h;
	int16_t xswap;
	int16_t yswap;

	if (!tab) {
		return;
	}

	if (tab->swap_tab) {
		tabs_paint_tab(tab->swap_tab);
	}

	c_window=windows+(tab->client->window);
	xswap=0;
	yswap=0;
	if ((tab->xswap)||(tab->yswap)) {
		memcpy(&tmp_window,c_window,sizeof(struct _window));
		if (tab->xswap<0) {
			tmp_window.w+=tab->xswap;
			tmp_window.mx+=tab->xswap;
			xswap=tab->xswap;
		} else {
			tmp_window.w-=tab->xswap;
			tmp_window.ox+=tab->xswap;
		}
		if (tab->yswap<0) {
			tmp_window.h+=tab->yswap;
			tmp_window.my+=tab->yswap;
			yswap=tab->yswap;
		} else {
			tmp_window.h-=tab->yswap;
			tmp_window.oy+=tab->yswap;
		}
		c_window=&tmp_window;
	}

	x=c_window->ox;
	y=c_window->oy;
	w=c_window->w;
	h=c_window->h;
	blit_fill(base,x,y,w,h,&tab->color,true);

	for(element=tab->widgets.list;element!=NULL;element=element->next) {
		widget=element->data;
		if ((!widget->visible)||(!widget->repaint_callback)) {
			continue;
		}
		final_x=x+widget->ox+xswap;
		final_y=y+widget->oy+yswap;
		if (!widget->h_fixed) {
			final_x+=tab->x_offset;
		}
		if (!widget->v_fixed) {
			final_y+=tab->y_offset;
		}
		widget->repaint_callback(widget,final_x,final_y,c_window);
	}
}


/**
 * Destroys a tab_scroll structure, freeing all its elements
 *
 * @param element The tab_scroll object to destroy
 *
 */

void tabs_free_tab_scroll(void *element) {

	struct _tab_scroll *object=element;

	if (!object) {
		return;
	}

	free(object);
}


void tabs_do_scroll() {

	struct minibus_list_element *element,*next;
	struct _tab_scroll *tsc;
	struct _tab *tab;
	int16_t dx,dy;

	for(element=tab_scroll.list;element!=NULL;element=next) {
		next=element->next;
		tsc=element->data;

		tab=tsc->tab;

		if (tsc->swap_tab==false) {

			// It's a classic scroll

			if (!(tsc->flags&TTOS_TAB_OFFSET_KEEP_HORIZONTAL)) {
				if (tsc->current_x>tsc->final_x) {
					dx=tsc->current_x-tsc->final_x;
				} else {
					dx=tsc->final_x-tsc->current_x;
				}
				if (dx>=TTOS_SCROLL_TAB_SPEED) {
					if ((tsc->final_x*tsc->current_x)<0) {
						tsc->current_x+=(tsc->current_x-tsc->final_x)/TTOS_SCROLL_TAB_SPEED;
					} else {
						tsc->current_x-=(tsc->current_x-tsc->final_x)/TTOS_SCROLL_TAB_SPEED;
					}
				} else {
					tsc->current_x=tsc->final_x;
				}
				tab->x_offset=tsc->current_x;
			}

			if (!(tsc->flags&TTOS_TAB_OFFSET_KEEP_VERTICAL)) {
				if (tsc->current_y>tsc->final_y) {
					dy=tsc->current_y-tsc->final_y;
				} else {
					dy=tsc->final_y-tsc->current_y;
				}
				if (dy>TTOS_SCROLL_TAB_SPEED) {
					if ((tsc->final_y*tsc->current_y)<0) {
						tsc->current_y+=(tsc->current_y-tsc->final_y)/TTOS_SCROLL_TAB_SPEED;
					} else {
						tsc->current_y-=(tsc->current_y-tsc->final_y)/TTOS_SCROLL_TAB_SPEED;
					}
				} else {
					tsc->current_y=tsc->final_y;
				}
				tab->y_offset=tsc->current_y;
			}

			if ((tsc->current_x==tsc->final_x)&&(tsc->current_y==tsc->final_y)) {
				tab->sensitive=true;
				if (tsc->event_port!=TTOS_NO_ANSWER) {
					minibus_start_message(write_conn,tsc->event_port,TTOS_NO_ANSWER);
					minibus_write_long(write_conn,TTOS_SR_SCROLL);
					minibus_write_short(write_conn,tab->tab_id);
					minibus_server_send(write_conn,tab->client->conn,false);
				}
				minibus_list_remove(element,true);
			}
		} else {
			// It's a tab swap

			struct _window *current_window;
			bool done;

			done=false;
			current_window=&(windows[tab->client->window]);
			switch(tsc->swap_mode) {
			case TTOS_TAB_SWAP_MODE_SCROLL_RIGHT:
				dx=-tab->xswap;
				if (dx>TTOS_SCROLL_TAB_SPEED) {
					tab->xswap+=dx/TTOS_SCROLL_TAB_SPEED;
					if (tab->swap_tab)
						tab->swap_tab->xswap+=dx/TTOS_SCROLL_TAB_SPEED;
				} else {
					tab->xswap++;
					if (tab->swap_tab)
						tab->swap_tab->xswap++;
				}
			break;
			case TTOS_TAB_SWAP_MODE_SCROLL_LEFT:
				dx=tab->xswap;
				if (dx>TTOS_SCROLL_TAB_SPEED) {
					tab->xswap-=dx/TTOS_SCROLL_TAB_SPEED;
					if (tab->swap_tab)
						tab->swap_tab->xswap-=dx/TTOS_SCROLL_TAB_SPEED;
				} else {
					tab->xswap--;
					if (tab->swap_tab)
						tab->swap_tab->xswap--;
				}
			break;
			case TTOS_TAB_SWAP_MODE_SCROLL_UP:
				dy=tab->yswap;
				if (dy>TTOS_SCROLL_TAB_SPEED) {
					tab->yswap-=dy/TTOS_SCROLL_TAB_SPEED;
					if (tab->swap_tab)
						tab->swap_tab->yswap-=dy/TTOS_SCROLL_TAB_SPEED;
				} else {
					tab->yswap--;
					if (tab->swap_tab)
						tab->swap_tab->yswap--;
				}
			break;
			case TTOS_TAB_SWAP_MODE_SCROLL_DOWN:
				dy=-tab->yswap;
				if (dy>TTOS_SCROLL_TAB_SPEED) {
					tab->yswap+=dy/TTOS_SCROLL_TAB_SPEED;
					if (tab->swap_tab)
						tab->swap_tab->yswap+=dy/TTOS_SCROLL_TAB_SPEED;
				} else {
					tab->yswap++;
					if (tab->swap_tab)
						tab->swap_tab->yswap++;
				}
			break;
			}
			if ((!tab->xswap)&&(!tab->yswap)) {
				tab->sensitive=true;
				if (tab->swap_tab) {
					tab->swap_tab->sensitive=true;
					if (tab->swap_tab->to_delete) {
						uid_release_uid(ttos_uids,(uint32_t)tab->swap_tab->tab_id);
						minibus_list_remove_data(&tab->client->tabs,tab->swap_tab,true);
					}
					tab->swap_tab=NULL;
				}
				if (tsc->event_port!=TTOS_NO_ANSWER) {
					minibus_start_message(write_conn,tsc->event_port,TTOS_NO_ANSWER);
					minibus_write_long(write_conn,TTOS_SR_TAB_SWAP_END);
					minibus_write_short(write_conn,tab->tab_id);
					minibus_server_send(write_conn,tab->client->conn,false);
				}
				minibus_list_remove(element,true);
			}
		}
	}
	main_refresh(NULL);
}


struct _tab_scroll *tabs_init_tab_scroll(struct _tab *tab, struct _tab *old_tab, int16_t new_x, int16_t new_y, uint32_t event_port, uint32_t flags, bool swap) {

	struct _tab_scroll *object;

	if (!tab) {
		return NULL;
	}

	object=malloc(sizeof(struct _tab_scroll));

	if (!object) {
		return NULL;
	}

	// if TTOS_SCROLL_SCREEN_SIZE_COORDS flag is set, the coordinates are "screens" instead of "pixels"
	if (flags&TTOS_TAB_OFFSET_SCREEN_SIZE_COORDS) {
		new_x*=base->w;
		new_y*=base->h;
	}

	object->original_x=object->current_x=tab->x_offset;
	object->original_y=object->current_y=tab->y_offset;
	if (!swap) {
		object->final_x=(flags&TTOS_TAB_OFFSET_KEEP_HORIZONTAL) ? tab->x_offset : new_x;
		object->final_y=(flags&TTOS_TAB_OFFSET_KEEP_VERTICAL) ? tab->y_offset : new_y;
		object->flags=flags;
		object->swap_tab=false;
	} else {
		if (old_tab)
			old_tab->sensitive=false;
		object->final_x=0;
		object->final_y=0;
		object->swap_mode=flags;
		object->swap_tab=true;
	}
	object->tab=tab;
	object->tab->sensitive=false;
	object->event_port=event_port;
	return (object);
}


/**
 * Sets the offset values for the specified tab
 *
 * Receives a message with three shorts and two longs: the tab UID, the new x and y offsets, the event port and the flags
 *
 * If FLAGS contains the TTOS_TAB_OFFSET_DO_SCROLL, it will do an animated scroll; if not, the offset will be set
 * instantaneously.
 */

uint32_t tabs_set_tab_offset(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	struct _client_data *client;
	struct minibus_data_element *data;
	struct _tab_scroll *scroll;
	uint16_t tab_id;
	int16_t new_x;
	int16_t new_y;
	uint32_t event_port;
	uint32_t flags;
	int32_t error;

	if (!connection) {
		return -1;
	}

	data=minibus_read_params(connection,"sssll",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	client=connection->data;
	tab_id=data[0].value16;

	if (!tabs_select_edit_tab(connection,tab_id)) {
		free(data);
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
		return -1;
	}

	new_x=data[1].value16;
	new_y=data[2].value16;
	event_port=data[3].value32;
	flags=data[4].value32;
	free(data);

	// If the tab to scroll is not the current visible tab, or if the client is not the currently
	// active client in the window, or we just don't want animation, just scroll the tab without animation

	if ((!(flags&TTOS_TAB_OFFSET_DO_SCROLL))||(windows[client->window].fg_client!=client)||(client->edit_tab!=client->show_tab)) {
		if (flags&TTOS_TAB_OFFSET_SCREEN_SIZE_COORDS) {
			new_x*=base->w;
			new_y*=base->h;
		}
		if (!(flags&TTOS_TAB_OFFSET_KEEP_HORIZONTAL))
			client->edit_tab->x_offset=new_x;
		if (!(flags&TTOS_TAB_OFFSET_KEEP_VERTICAL))
			client->edit_tab->y_offset=new_y;
		other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
		// send also the scroll event
		if (event_port!=TTOS_NO_ANSWER) {
			minibus_start_message(write_conn,event_port,TTOS_NO_ANSWER);
			minibus_write_long(write_conn,TTOS_SR_SCROLL);
			minibus_write_short(write_conn,client->edit_tab->tab_id);
			minibus_server_send(write_conn,connection,false);
		}
		return 0;
	}

	scroll=tabs_init_tab_scroll(client->edit_tab,NULL,new_x,new_y,event_port,flags,false);

	if(!scroll) {
		other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
		return -1;
	}

	minibus_list_add(&tab_scroll,scroll);
	timer_start_scroll();

	other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
	return 0;
}

