/*
 * Copyright 2011 Raster Software Vigo
 *
 * This file is part of ttos
 *
 * ttos is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * ttos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * timer.c
 *
 *  Created on: 10/10/2010
 *      Author: raster
 */

#include <stdlib.h>
#include <sys/time.h>

#include "minibus/minibus.h"
#include "global_vars.h"
#include "timer.h"
#include "ttos_lib/ttos_structs.h"
#include "other.h"
#include "tabs.h"
#include "uid.h"


/**
 * Returns the current time in millisecond since epoch
 */

uint64_t timer_get_time() {

	struct timeval tv;
	uint64_t time;

	gettimeofday(&tv,NULL);

	time=1000*((uint64_t)tv.tv_sec);
	time+=(((uint64_t)tv.tv_usec)/1000);
	return (time);
}

void timer_do_actions(enum ttos_timer_type timer_t,uint32_t port,uint16_t uid, struct minibus_connection *conn) {

	switch(timer_t) {
	case TTOS_TIMER_USER_DEFINED:
		minibus_start_message(write_conn,port,TTOS_NO_ANSWER);
		minibus_write_long(write_conn,TTOS_SR_TIMER);
		minibus_write_short(write_conn,uid);
		minibus_server_send(write_conn,conn,false);
	break;
	case TTOS_TIMER_SCROLL_TABS:
		tabs_do_scroll();
		if (tab_scroll.list) {
			timer_start_scroll();
		}
	break;
	case TTOS_TIMER_DOUBLE_CLICK:
		timer_check_event(TTOS_TIMER_DOUBLE_CLICK,true,false,0);
	break;
	case TTOS_TIMER_LONG_CLICK:
		if (window_manager) {
			large_touch_made=true;
			minibus_start_message(write_conn,TTOS_SYSTEM_PORT_2,TTOS_NO_ANSWER);
			minibus_write_long(write_conn,TTOS_SR_TOUCH_EVENT);
			minibus_write_byte(write_conn,TTOS_TS_MASTER);
			minibus_write_short(write_conn,0);
			minibus_write_short(write_conn,0);
			minibus_write_short(write_conn,0);
			minibus_write_short(write_conn,0);
			minibus_write_short(write_conn,0);
			minibus_write_short(write_conn,0);
			minibus_server_send(write_conn,window_manager->conn,false);
		}
	break;
	}
}

int32_t timer_timeout_callback(struct minibus_master *object,uint32_t timeout,void *data) {

	uint64_t currtime;
	uint32_t port;
	uint16_t uid;
	struct _timer *timer;
	struct minibus_connection *conn;
	enum ttos_timer_type timer_t;

	currtime=timer_get_time();

	if(!timers.list) {
		return -1;
	}

	while(timers.list) {
		timer=(struct _timer *)timers.list->data;
		if (timer->instant>currtime) {
			break;
		}
		timer_t=timer->type;
		uid=timer->uid;
		port=timer->port;
		conn=timer->conn;

		if ((timer->period)&&(timer->type!=TTOS_TIMER_SCROLL_TABS)) {
			//minibus_list_incref(timers.list); // increment reference counter to avoid the memory to be released in the follow REMOVE
			minibus_list_remove(timers.list,false);
			timer->instant+=timer->period; // this avoids accumulative errors
			minibus_list_add_sorted(&timers,timer);
		} else {
			minibus_list_remove(timers.list,true);
		}
		timer_do_actions(timer_t,port,uid,conn);
	}

	if (!timers.list) {
		return -1;
	}
	timer=(struct _timer *)timers.list->data;
	return((int32_t)((timer->instant)-currtime));

}

/**
 * Initializates a timer structure
 *
 * @param type The type: an user-defined timer, or a system timer
 * @param connection If it's an user-defined timer, the connection which asked for it
 * @param uid The timer's UID
 * @param port If it's an user-defined timer, the port to which send the event
 * @param period The period for this timer, or 0 for a one-shot timer
 * @param instant How many time from now has to be triggered for first time
 *
 * @return A pointer to an initialized timer struct, or NULL if there's an error
 *
 */

struct _timer *timer_init_timer(enum ttos_timer_type type,struct minibus_connection *connection,uint16_t uid,uint32_t port,uint64_t period, uint64_t instant) {

	struct _timer *timer;

	if ((!connection)&&(type==TTOS_TIMER_USER_DEFINED)) {
		return NULL;
	}

	timer=(struct _timer *)malloc(sizeof(struct _timer));
	if (!timer) {
		return NULL;
	}

	timer->type=type;
	timer->conn=connection;
	timer->uid=uid;
	timer->port=port;
	timer->period=period;
	timer->instant=instant;
	return (timer);
}


/**
 * Compare function for TIMER list
 *
 * @param timer1
 * @param timer2
 * @return
 */

bool timer_compare(void *timer1, void *timer2) {

	struct _timer *t1,*t2;

	t1=timer1;
	t2=timer2;

	if (!t1) {
		return false;
	}
	if (!t2) {
		return true;
	}
	return ((t1->instant)<=(t2->instant));

}

/**
 * Creates a new timer.
 *
 * Receives a message with the port to which send the timeout event,
 * the period in milliseconds (long) and if it is a one-shot timer or not (byte)
 */

uint32_t timer_new_timer(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	struct minibus_data_element *data;
	struct _timer *timer;
	int32_t error;
	uint64_t period;
	uint8_t oneshot;
	uint64_t ctime;
	uint32_t timer_uid;
	uint32_t event_port;

	data=minibus_read_params(connection,"llb",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}

	event_port=data[0].value32;
	period=(uint64_t)data[1].value32;
	oneshot=data[2].value8;
	free(data);

	timer_uid=uid_get_uid(ttos_uids,TTOS_TIMER_UID_MASK);

	ctime=timer_get_time();
	timer=timer_init_timer(TTOS_TIMER_USER_DEFINED,connection,(uint16_t)(timer_uid&0x0000FFFF),event_port,oneshot ? 0 : period, period+ctime);

	if (!timer) {
		other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
		return -1;
	}

	minibus_list_add_sorted(&timers,timer);
	other_send_uid(connection,answer_port,timer_uid);
	return 0;
}


/**
 * Checks if there's a timer event of the specified type, returning the result. Also allows to remove it, and insert another one
 *
 * @param type The event type to check
 * @param remove If true, the event will be removed from the queue
 * @param add If true, a new event will be added, only if currently there are no more events of that type
 * @param period The period for the new event, in case it has to be added
 *
 * @return True if now there was an event of the specified type before removing it or adding another; False if not
 */

bool timer_check_event(enum ttos_timer_type type, bool remove, bool add, uint64_t period) {

	struct minibus_list_element *element,*element2;
	struct _timer *timer;
	bool exists;
	uint64_t ctime;

	exists=false;
	for(element=timers.list;element;element=element2) {
		element2=element->next;
		timer=element->data;
		if (timer->type==type) { //There's a click timer started; kill it
			if (remove) {
				minibus_list_remove(element,true);
			} else {
				exists=true;
			}
		}
	}

	if ((!exists)&&(add)) {
		ctime=timer_get_time();
		timer=timer_init_timer(type,NULL,0,0,0,period+ctime);
		if (!timer) {
			return (false);
		}
		minibus_list_add_sorted(&timers,timer);
	}
	return(exists);
}


/**
 * Starts an internal timer for scrolling windows
 */

void timer_start_scroll() {

	// If there's no SCROLL event, add it with a period of TTOS_SCROLL_TAB_TIMER
	timer_check_event(TTOS_TIMER_SCROLL_TABS,false,true,TTOS_SCROLL_TAB_TIMER);

}

/**
 * Deletes a timer.
 *
 * Receives a message with the timer uid (short) to delete
 */

uint32_t timer_delete_timer(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	struct minibus_data_element *data;
	struct minibus_list_element *element,*next;
	struct _timer *timer;
	int16_t uid;
	int32_t error;

	data=minibus_read_params(connection,"s",&error);

	if(!data) {
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}
	uid=data[0].value16;
	free(data);

	for(element=timers.list;element!=NULL;element=next) {
		next=element->next;
		timer=element->data;
		if ((timer->uid==uid)&&(timer->conn==connection)) {
			uid_release_uid(ttos_uids,(uint32_t)(uid));
			minibus_list_remove(element,true);
		}
	}
	return 0;
}

void timer_destroy(void *object) {

	free(object);

}
