/*
 * Copyright 2011 Raster Software Vigo
 *
 * This file is part of ttos
 *
 * ttos is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * ttos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * ttos_lib.c
 *
 *  Created on: 12/10/2010
 *      Author: raster
 */

/** @file */

#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <libintl.h>
#include <errno.h>
#include "ttos.h"

#define _(string) gettext (string)

/**
 * Initalizates a connection with the window system and returns a pointer to a connection object
 *
 * @param flags Connection flags. If TTOS_INIT_ASK_WM, this app asks to be the WindowMaster. Only one app can be it
 * @param argc The argc parameter from the MAIN function
 * @param argv The argv parameter from the MAIN function
 *
 * @return A pointer to a server object, or NULL in case of failure
 */

struct ttos_object *ttos_init_system (uint32_t flags,int argc,char **argv) {

	struct minibus_connection *conn;
	struct minibus_master *queue;
	struct ttos_object *master;
	Cstring filename;
	uint32_t error;

	master=(struct ttos_object *)malloc(sizeof(struct ttos_object));
	if (!master) {
		return (NULL);
	}

	master->cb_port0=NULL;
	master->cb_port0_data=NULL;
	master->exit_cb=NULL;
	master->exit_cb_data=NULL;
	master->new_client=NULL;
	master->new_client_data=NULL;
	master->destroyed_client=NULL;
	master->destroyed_client_data=NULL;
	master->fg=false;
	master->shcounter=0;
	master->uid=0;

	queue=minibus_init("/tmp/ttos_queue",false);
	if (!queue) {
		free(master);
		return NULL;
	}

	conn=minibus_get_first_connection(queue);
	if (!conn) {
		free(master);
		minibus_release(queue);
		return NULL;
	}

	master->conn=conn;
	master->queue=queue;
	// system messages sent to this port will always be processed ASAP
	minibus_add_callback(queue,TTOS_SYSTEM_PORT_1,ttos_main_callback,master);
	// system messages sent to this port can be delayed when doing a minibus_wait_for_port call
	minibus_add_callback(queue,TTOS_SYSTEM_PORT_2,ttos_main_callback,master);

	minibus_start_message(conn,TTOS_SM_INIT,TTOS_RESERVED_PORTS);
	minibus_write_long(conn,flags);
	if (argc) {
		free(ttos_get_upper_path(argv[0],&filename));
		minibus_write_string(conn,filename);
	} else {
		minibus_write_string(conn,"");
	}
	minibus_send(conn);

	error=TTOS_ERROR_NO_ERROR;
	if ((0!=ttos_wait_for_error_2(master,TTOS_RESERVED_PORTS,&error))||(error!=TTOS_ERROR_NO_ERROR)) {
		free(master);
		minibus_release(queue);
		return NULL;
	}

	while(!master->uid) {
		minibus_wait_for_port(master->conn,TTOS_SYSTEM_PORT_1,true);
	}

	return (master);
}

/**
 * Fully closes a connection to the window system, freeing all its memory.
 *
 * @param object The server object
 */


void ttos_destroy_system(struct ttos_object *object) {

	if (!object) {
		return;
	}

	minibus_remove_callback(object->queue,TTOS_SYSTEM_PORT_1);
	minibus_remove_callback(object->queue,TTOS_SYSTEM_PORT_2);

	minibus_release(object->queue);

	free(object);

}

/**
 * Defines the windows in the system
 *
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param nwindows Number of windows to define
 * @param windows An array with the geometry of each window
 *
 * @return A TTOS_SYSTEM_ERRORS value
 */

int32_t ttos_define_windows(struct ttos_object *master,uint32_t answer_port,uint8_t nwindows,
		struct ttos_window_geometry windows[]) {

	uint8_t loop;
	uint32_t answer_port2;

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;

	minibus_start_message(master->conn,TTOS_SM_DEFINE_WINDOWS,answer_port2);
	minibus_write_long(master->conn,nwindows);

	for(loop=0;loop<nwindows;loop++) {
		minibus_write_short(master->conn,windows[loop].x);
		minibus_write_short(master->conn,windows[loop].y);
		minibus_write_short(master->conn,windows[loop].w);
		minibus_write_short(master->conn,windows[loop].h);
		minibus_write_byte(master->conn,windows[loop].visible);
	}
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}

	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Sends to a client the termination signal. Used by the window manager when shutting down the system
 *
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param app_uid The applicatin UID
 *
 * @return A TTOS_SYSTEM_ERRORS value
 */

int32_t ttos_close_client(struct ttos_object *master, uint32_t answer_port, uint32_t app_uid) {

	uint32_t answer_port2;

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;

	minibus_start_message(master->conn,TTOS_SM_CLOSE_CLIENT,answer_port2);
	minibus_write_long(master->conn,app_uid);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}

/**
 * Changes a window's visibility status
 *
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param window The window number to show/hide
 * @param visible True to make the window visible; False to make it invisible
 * @return The received error, or TTOS_ERROR_ASYNC_RESPONSE if an asynchronous answer have to be waited
 */

int32_t ttos_set_window_visibility(struct ttos_object *master,uint32_t answer_port,uint8_t window, bool visible) {

	uint32_t answer_port2;

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;

	minibus_start_message(master->conn,TTOS_SM_WINDOW_SET_VISIBLE,answer_port2);
	minibus_write_byte(master->conn,window);
	minibus_write_byte(master->conn,visible);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Creates a new tab for this application
 *
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 *
 * @return The received error, the TAB's UID, or TTOS_ERROR_ASYNC_RESPONSE if an asynchronous answer have to be waited
 */

int32_t ttos_new_tab(struct ttos_object *master,uint32_t answer_port) {

	uint32_t answer_port2;

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_NEW_TAB,answer_port2);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Destroys a tab for this application
 *
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param tab_uid The tab's UID
 *
 * @return The received error, or TTOS_ERROR_ASYNC_RESPONSE if an asynchronous answer have to be waited
 */

int32_t ttos_destroy_tab(struct ttos_object *master,uint32_t answer_port,uint32_t tab_uid, uint32_t flags) {

	uint32_t answer_port2;

	if ((tab_uid&0xF0000000)!=TTOS_TAB_UID_MASK) {
		return (TTOS_ERROR_INVALID_UID);
	}

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_DESTROY_TAB,answer_port2);
	minibus_write_short(master->conn,(uint16_t)(tab_uid&0x0000FFFF));
	minibus_write_long(master->conn,flags);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}

/**
 * Sets the tab's offset for the mobile widgets
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @return The received error, the TAB's UID, or TTOS_ERROR_ASYNC_RESPONSE if an asynchronous answer have to be waited
 */

int32_t ttos_get_visible_tab(struct ttos_object *master, uint32_t answer_port) {

	uint32_t answer_port2;

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_GET_CURRENT_TAB,answer_port2);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Changes the currently visible tab for this application
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param tab_uid The tab's UID
 * @param swap_mode One of the TTOS_TAB_SWAP_MODE_ defines, specifying how to move from the current tab to the new one
 * @param cb_port Port to which a TTOS_SR_TAB_SWAP_END event will be sent when the swapping process ended
 *
 * @return The received error, or TTOS_ERROR_ASYNC_RESPONSE if an asynchronous answer have to be waited
 */

int32_t ttos_set_visible_tab(struct ttos_object *master,uint32_t answer_port,uint32_t tab_uid,uint8_t swap_mode, uint32_t cb_port) {

	uint32_t answer_port2;

	if ((tab_uid&0xF0000000)!=TTOS_TAB_UID_MASK) {
		return (TTOS_ERROR_INVALID_UID);
	}

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_SELECT_TAB,answer_port2);
	minibus_write_short(master->conn,(uint16_t)(tab_uid&0x0000FFFF));
	minibus_write_byte(master->conn,swap_mode);
	minibus_write_long(master->conn,cb_port);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Sets the tab's background color
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param tab_uid The tab's UID
 * @param r The red component
 * @param g The red component
 * @param b The red component
 *
 * @return 0 if everything went fine, -1 if there was an error
 */

int32_t ttos_set_tab_bg(struct ttos_object *master,uint32_t answer_port,uint32_t tab_uid,struct ttos_color *color) {

	uint32_t answer_port2;

	if ((tab_uid&0xF0000000)!=TTOS_TAB_UID_MASK) {
		return (TTOS_ERROR_INVALID_UID);
	}

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_SET_TAB_BG_COLOR,answer_port2);
	minibus_write_short(master->conn,(uint16_t)(tab_uid&0x0000FFFF));
	if (!color) {
		minibus_write_byte(master->conn,0);
		minibus_write_byte(master->conn,0);
		minibus_write_byte(master->conn,0);
		minibus_write_byte(master->conn,0);
	} else {
		minibus_write_byte(master->conn,color->r);
		minibus_write_byte(master->conn,color->g);
		minibus_write_byte(master->conn,color->b);
		minibus_write_byte(master->conn,color->a);
	}
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Sets the tab's size
 *
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param tab_uid The tab's UID
 * @param x The start X coordinate
 * @param y The start Y coordinate
 * @param w The width
 * @param h The height
 * @param force If true, forces the specified size, even if its bigger or smaller than the current one. If not, it will be set only if it's bigger than the minimun required
 *
 * @return 0 if everything went fine, -1 if there was an error
 */

int32_t ttos_set_tab_size(struct ttos_object *master,uint32_t answer_port,uint32_t tab_uid,int16_t x, int16_t y, uint16_t w, uint16_t h, bool force) {

	uint32_t answer_port2;

	if ((tab_uid&0xF0000000)!=TTOS_TAB_UID_MASK) {
		return (TTOS_ERROR_INVALID_UID);
	}

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_SET_TAB_SIZE,answer_port2);
	minibus_write_short(master->conn,(uint16_t)(tab_uid&0x0000FFFF));
	minibus_write_short(master->conn,x);
	minibus_write_short(master->conn,y);
	minibus_write_short(master->conn,w);
	minibus_write_short(master->conn,h);
	minibus_write_byte(master->conn,force);

	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Set extra margins for each widget in a tab, allowing to add extra padding when doing automatic scroll
 *
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param tab_uid The tab's UID
 * @param margin_up The value for the upper margin
 * @param margin_down The value for the lower margin
 * @param margin_left The value for the left margin
 * @param margin_right The value for the right margin
 * @return
 */

int32_t ttos_set_tab_margins(struct ttos_object *master, uint32_t answer_port, uint32_t tab_uid, uint16_t margin_up,
		uint16_t margin_down, uint16_t margin_left, uint16_t margin_right) {

	uint32_t answer_port2;

	if ((tab_uid&0xF0000000)!=TTOS_TAB_UID_MASK) {
		return (TTOS_ERROR_INVALID_UID);
	}

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_SET_TAB_MARGINS,answer_port2);
	minibus_write_short(master->conn,(uint16_t)(tab_uid&0x0000FFFF));
	minibus_write_short(master->conn,margin_up);
	minibus_write_short(master->conn,margin_down);
	minibus_write_short(master->conn,margin_left);
	minibus_write_short(master->conn,margin_right);

	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Repaints the screen
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 *
 * @return The received error, or TTOS_ERROR_ASYNC_RESPONSE if an asynchronous answer have to be waited
 */

int32_t ttos_repaint_screen(struct ttos_object *master,uint32_t answer_port) {

	uint32_t answer_port2;

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_REPAINT_SCREEN,answer_port2);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Creates a new timer, which can be one-shot or periodic
 *
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param timer_uid The desired UID for the new timer
 * @param timer_period The timer's period in milliseconds
 * @param cb_port The port to which send the timer signal
 * @param periodic If false, the timer will be one-shot; if false, will be periodic
 *
 * @return 0 if everything went fine, -1 if there was an error
 */


int32_t ttos_create_timer(struct ttos_object *master, uint32_t answer_port, uint32_t timer_period, uint32_t cb_port, bool periodic) {

	uint32_t answer_port2;

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_NEW_TIMER,answer_port2);
	minibus_write_long(master->conn,cb_port);
	minibus_write_long(master->conn,timer_period);
	minibus_write_byte(master->conn,periodic==false ? 1 : 0);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Deletes a previously set timer
 *
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param timer_uid The timer's UID
 *
 * @return 0 if everything went fine, -1 if there was an error
 */

int32_t ttos_delete_timer (struct ttos_object *master, uint32_t answer_port, uint32_t timer_uid) {

	uint32_t answer_port2;
	uint16_t uid;

	if ((timer_uid&0xF0000000)!=TTOS_TIMER_UID_MASK) {
		return (TTOS_ERROR_INVALID_UID);
	}

	uid=(uint16_t) (timer_uid&0x0000FFFF);
	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_DELETE_TIMER,answer_port2);
	minibus_write_short(master->conn,uid);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Sets the tab offset, allowing to do it instantaneously or with a smooth scroll
 *
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param tab_uid The tab to set
 * @param new_x New X offset
 * @param new_y New Y offset
 * @param cb_port The port into which receive the event when the scroll ends
 * @param flags several flags about how to set the offset (with/without scrool; keeping the old tab below...)
 * @return a generic error
 */

int32_t ttos_set_tab_offset (struct ttos_object *master,uint32_t answer_port, uint32_t tab_uid, int16_t new_x, int16_t new_y, uint32_t cb_port, uint32_t flags) {

	uint32_t answer_port2;

	if ((tab_uid&0xF0000000)!=TTOS_TAB_UID_MASK) {
		return (TTOS_ERROR_INVALID_UID);
	}

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_SET_TAB_OFFSET,answer_port2);
	minibus_write_short(master->conn,(uint16_t)(tab_uid&0x0000FFFF));
	minibus_write_short(master->conn,new_x);
	minibus_write_short(master->conn,new_y);
	minibus_write_long(master->conn,cb_port);
	minibus_write_long(master->conn,flags);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Moves an application to foreground
 *
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param app_uid The application's UID
 *
 * @return The received error, or TTOS_ERROR_ASYNC_RESPONSE if an asynchronous answer have to be waited
 */

int32_t ttos_set_fg_app (struct ttos_object *master, uint32_t answer_port, uint32_t app_uid) {

	uint32_t answer_port2;

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;

	minibus_start_message(master->conn,TTOS_SM_MOVE_CLIENT_TO_FG,answer_port2);
	minibus_write_long(master->conn,app_uid);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Sets the window to be used by an application
 *
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param app_uid The application's UID
 * @param window_number The window to be used by the application
 * @return The received error, or TTOS_ERROR_ASYNC_RESPONSE if an asynchronous answer have to be waited
 */

int32_t ttos_set_client_window (struct ttos_object *master, uint32_t answer_port, uint32_t app_uid, uint8_t window_number) {

	uint32_t answer_port2;

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;

	minibus_start_message(master->conn,TTOS_SM_SET_CLIENT_WINDOW,answer_port2);
	minibus_write_long(master->conn,app_uid);
	minibus_write_byte(master->conn,window_number);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Changes the visibility status of a widget
 *
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param widget The widget's UID
 * @param visible The new visibility status
 *
 * @return An error message
 */


uint32_t ttos_set_visibility (struct ttos_object *master, uint32_t answer_port, uint32_t widget_uid, bool visible) {

	uint32_t answer_port2;

	if ((widget_uid&0xF0000000)!=TTOS_WIDGET_UID_MASK) {
		return (TTOS_ERROR_INVALID_UID);
	}

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_SET_WIDGET_VISIBILITY,answer_port2);
	minibus_write_long(master->conn,widget_uid);
	minibus_write_byte(master->conn,visible);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Changes the screen position of a widget
 *
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param widget The widget's UID
 * @param x The new X coordinate
 * @param y The new Y coordinate
 *
 * @return An error message
 */


uint32_t ttos_set_widget_position (struct ttos_object *master, uint32_t answer_port, uint32_t widget_uid, int16_t x, int16_t y) {

	uint32_t answer_port2;

	if ((widget_uid&0xF0000000)!=TTOS_WIDGET_UID_MASK) {
		return (TTOS_ERROR_INVALID_UID);
	}

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_SET_WIDGET_POSITION,answer_port2);
	minibus_write_long(master->conn,widget_uid);
	minibus_write_short(master->conn,x);
	minibus_write_short(master->conn,y);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Places a picture in a tab, and sets a callback  (if desired) to receive click and drag events made over it
 *
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param tab The tab into which place the picture
 * @param x The x coordinate
 * @param y The y coordinate
 * @param filename The path and filename of the picture
 * @param flags Picture flags
 * @param cb_port The port which will receive the events (0 for no sensitivity)
 *
 * @return An UID or an error message
 */

uint32_t ttos_new_picture (struct ttos_object *master, uint32_t answer_port, uint32_t tab_uid, int16_t x, int16_t y,
						 Cstring filename, uint32_t flags, uint32_t cb_port) {

	uint32_t answer_port2;

	if ((tab_uid&0xF0000000)!=TTOS_TAB_UID_MASK) {
		return (TTOS_ERROR_INVALID_UID);
	}

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_NEW_WIDGET_PICTURE,answer_port2);
	minibus_write_short(master->conn,(uint16_t)(tab_uid&0x0000FFFF));
	minibus_write_short(master->conn,x);
	minibus_write_short(master->conn,y);
	minibus_write_long(master->conn,flags);
	minibus_write_long(master->conn,cb_port);
	minibus_write_string(master->conn,filename);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Replaces the current picture in a picture widget with a new picture
 *
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param picture_uid The widget to which replace the picture
 * @param filename The new picture to use
 * @return An error
 */


uint32_t ttos_set_picture (struct ttos_object *master, uint32_t answer_port, uint32_t picture_uid, Cstring filename) {

	uint32_t answer_port2;

	if ((picture_uid&0xF0000000)!=TTOS_WIDGET_UID_MASK) {
		return (TTOS_ERROR_INVALID_UID);
	}

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_SET_PICTURE,answer_port2);
	minibus_write_long(master->conn,picture_uid);
	minibus_write_string(master->conn,filename);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Places a label in a tab
 *
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param tab The tab into which place the picture
 * @param x The x coordinate
 * @param y The y coordinate
 * @param w The width
 * @param h The height
 * @param text The text to set in the label
 * @param color A pointer to a color structure, or NULL for default color (black)
 * @param flags Label flags
 * @param cb_port The port which will receive the events (0 for no sensitivity)
 *
 * @return a TTOS_SYSTEM_ERRORS value or an UID
 */

uint32_t ttos_new_label (struct ttos_object *master, uint32_t answer_port, uint32_t tab_uid, int16_t x, int16_t y,
						 int16_t w, int16_t h, Cstring text, struct ttos_color *color, uint32_t flags, uint32_t cb_port) {

	uint32_t answer_port2;
	uint8_t r,g,b;

	if ((tab_uid&0xF0000000)!=TTOS_TAB_UID_MASK) {
		return (TTOS_ERROR_INVALID_UID);
	}

	if (!color) {
		r=g=b=0;
	} else {
		r=color->r;
		g=color->g;
		b=color->b;
	}

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_NEW_WIDGET_LABEL,answer_port2);
	minibus_write_short(master->conn,(uint16_t)(tab_uid&0x0000FFFF));
	minibus_write_short(master->conn,x);
	minibus_write_short(master->conn,y);
	minibus_write_short(master->conn,w);
	minibus_write_short(master->conn,h);
	minibus_write_long(master->conn,flags);
	minibus_write_long(master->conn,cb_port);
	minibus_write_string(master->conn,text);
	minibus_write_byte(master->conn,r);
	minibus_write_byte(master->conn,g);
	minibus_write_byte(master->conn,b);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Places a color area in a tab
 *
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param tab The tab into which place the picture
 * @param x The x coordinate
 * @param y The y coordinate
 * @param w The width
 * @param h The height
 * @param color A pointer to a color structure, or NULL for default color (black)
 * @param flags Label flags
 * @param cb_port The port which will receive the events (0 for no sensitivity)
 *
 * @return An UID or an error message
 */

uint32_t ttos_new_color_area (struct ttos_object *master, uint32_t answer_port, uint32_t tab_uid, int16_t x, int16_t y,
						 int16_t w, int16_t h, struct ttos_color *color, uint32_t flags, uint32_t cb_port) {

	uint32_t answer_port2;
	uint8_t r,g,b,a;

	if ((tab_uid&0xF0000000)!=TTOS_TAB_UID_MASK) {
		return (TTOS_ERROR_INVALID_UID);
	}

	if (!color) {
		r=g=b=0;
		a=255;
	} else {
		r=color->r;
		g=color->g;
		b=color->b;
		a=color->a;
	}

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_NEW_WIDGET_COLOR_AREA,answer_port2);
	minibus_write_short(master->conn,(uint16_t)(tab_uid&0x0000FFFF));
	minibus_write_short(master->conn,x);
	minibus_write_short(master->conn,y);
	minibus_write_short(master->conn,w);
	minibus_write_short(master->conn,h);
	minibus_write_long(master->conn,flags);
	minibus_write_long(master->conn,cb_port);
	minibus_write_byte(master->conn,r);
	minibus_write_byte(master->conn,g);
	minibus_write_byte(master->conn,b);
	minibus_write_byte(master->conn,a);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Places a button in a tab
 *
 * @param master The current connection with the window server
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param tab The tab into which place the picture
 * @param x The x coordinate
 * @param y The y coordinate
 * @param w The width
 * @param h The height
 * @param stock_button An stock button, or TTOS_NO_STOCK_BUTTON (0) for an user-defined button
 * @param picture The picture path to set in the button
 * @param text The text to set in the button
 * @param color A pointer to a color structure with the desired text color, or NULL for default (white) color
 * @param bgcolor A pointer to a color structure with the desired background color, or NULL for default (black) color
 * @param flags Label flags
 * @param cb_port The port which will receive the events (0 for no sensitivity)
 *
 * @return An UID or an error message
 */

uint32_t ttos_new_button (struct ttos_object *master, uint32_t answer_port, uint32_t tab_uid, int16_t x, int16_t y,
						 int16_t w, int16_t h, enum ttos_stock_button stock_button, Cstring picture,Cstring text,
						 struct ttos_color *color, struct ttos_color *bgcolor, uint32_t flags, uint32_t cb_port) {

	uint32_t answer_port2;
	uint8_t r,g,b;
	uint8_t br,bg,bb,ba;

	if ((tab_uid&0xF0000000)!=TTOS_TAB_UID_MASK) {
		return (TTOS_ERROR_INVALID_UID);
	}

	if (!color) {
		r=g=b=0xFF;
	} else {
		r=color->r;
		g=color->g;
		b=color->b;
	}
	if (!bgcolor) {
		br=bg=bb=0x0F;
		ba=0xC0;
	} else {
		br=bgcolor->r;
		bg=bgcolor->g;
		bb=bgcolor->b;
		ba=bgcolor->a;
	}

	switch (stock_button) {
	case TTOS_OK_BUTTON:
		text=_("OK");
		picture="ok.tga";
	break;
	case TTOS_CANCEL_BUTTON:
		text=_("Cancel");
		picture="cancel.tga";
	break;
	case TTOS_EXIT_BUTTON:
		text=_("Exit");
		picture="exit.tga";
	break;
	case TTOS_LEFT_BUTTON:
		text=_("Back");
		picture="arrow_left.tga";
	break;
	case TTOS_RIGHT_BUTTON:
		text=_("Forward");
		picture="arrow_right.tga";
	break;
	case TTOS_UP_BUTTON:
		text=_("Up");
		picture="arrow_up.tga";
	break;
	case TTOS_INTRO_BUTTON:
		text=_("Intro");
		picture="arrow_intro.tga";
	break;
	case TTOS_DOWN_BUTTON:
		text=_("Down");
		picture="arrow_down.tga";
	break;
	default:
	break;
	}

	if ((!text) || (flags&TTOS_WIDGET_BUTTON_NO_TEXT)) {
		text="";
	}

	if (!picture) {
		picture="";
	}

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_NEW_WIDGET_BUTTON,answer_port2);
	minibus_write_short(master->conn,(uint16_t)(tab_uid&0x0000FFFF));
	minibus_write_short(master->conn,x);
	minibus_write_short(master->conn,y);
	minibus_write_short(master->conn,w);
	minibus_write_short(master->conn,h);
	minibus_write_long(master->conn,flags|TTOS_WIDGET_CLICK_EVENT);
	minibus_write_long(master->conn,cb_port);
	minibus_write_string(master->conn,picture);
	minibus_write_string(master->conn,text);
	minibus_write_byte(master->conn,r);
	minibus_write_byte(master->conn,g);
	minibus_write_byte(master->conn,b);
	minibus_write_byte(master->conn,br);
	minibus_write_byte(master->conn,bg);
	minibus_write_byte(master->conn,bb);
	minibus_write_byte(master->conn,ba);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Allows to replace the text in a label
 *
 * @param master The connection to the compositor
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param widget_uid The widget's UID
 * @param text The new text
 * @param mask The text flags
 * @return
 */


uint32_t ttos_set_label_text(struct ttos_object *master, uint32_t answer_port, uint32_t widget_uid, Cstring text,uint32_t mask) {

	uint32_t answer_port2;

	if ((widget_uid&0xF0000000)!=TTOS_WIDGET_UID_MASK) {
		return (TTOS_ERROR_INVALID_UID);
	}

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_SET_LABEL_TEXT,answer_port2);
	minibus_write_long(master->conn,widget_uid);
	minibus_write_long(master->conn,mask);
	minibus_write_string(master->conn,text);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Creates a new drawable
 *
 * @param master The connection to the compositor
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param tab_uid The tab into which create the drawable
 * @param x The X coordinate where to put the drawable
 * @param y The Y coordinate where to put the drawable
 * @param w The width
 * @param h The height
 * @param flags Flags for this drawable
 * @param cb_port The port which will receive the events (0 for no sensitivity)
 * @return A widget UID, or an Error
 */

uint32_t ttos_new_drawable (struct ttos_object *master, uint32_t answer_port, uint32_t tab_uid, int16_t x, int16_t y,
						 int16_t w, int16_t h, uint32_t flags, uint32_t cb_port) {

	uint32_t answer_port2;

	if ((tab_uid&0xF0000000)!=TTOS_TAB_UID_MASK) {
		return (TTOS_ERROR_INVALID_UID);
	}

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_NEW_WIDGET_DRAWABLE,answer_port2);
	minibus_write_short(master->conn,(uint16_t)(tab_uid&0x0000FFFF));
	minibus_write_short(master->conn,x);
	minibus_write_short(master->conn,y);
	minibus_write_short(master->conn,w);
	minibus_write_short(master->conn,h);
	minibus_write_long(master->conn,0);
	minibus_write_long(master->conn,flags);
	minibus_write_long(master->conn,cb_port);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Creates a new drawable, using a sharemem object as memory. Useful for animations thanks to its faster update.
 *
 * @param master The connection to the compositor
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param tab_uid The tab into which create the drawable
 * @param x The X coordinate where to put the drawable
 * @param y The Y coordinate where to put the drawable
 * @param sharemem The sharemem object to use
 * @param flags Flags for this drawable
 * @param cb_port The port which will receive the events (0 for no sensitivity)
 * @return A widget UID, or an Error
 */

uint32_t ttos_new_drawable_from_sharemem (struct ttos_object *master, uint32_t answer_port, uint32_t tab_uid, int16_t x,
						int16_t y, struct ttos_sharemem_object * sharemem, uint32_t flags, uint32_t cb_port) {

	uint32_t answer_port2;

	if ((tab_uid&0xF0000000)!=TTOS_TAB_UID_MASK) {
		return (TTOS_ERROR_INVALID_UID);
	}

	if (!master || !sharemem) {
		return TTOS_ERROR_INVALID_OBJECT;
	}

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_NEW_WIDGET_DRAWABLE,answer_port2);
	minibus_write_short(master->conn,(uint16_t)(tab_uid&0x0000FFFF));
	minibus_write_short(master->conn,x);
	minibus_write_short(master->conn,y);
	minibus_write_short(master->conn,0);
	minibus_write_short(master->conn,0);
	minibus_write_long(master->conn,sharemem->uid);
	minibus_write_long(master->conn,flags);
	minibus_write_long(master->conn,cb_port);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}



/**
 * Updates a drawable with the pixel data specified, using the main queue. This is the only way of update a drawable
 * not using a sharemem object.
 *
 * @param master The connection to the compositor
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param widget_uid The drawable widget to update
 * @param offset Offset (in pixels) to the start point
 * @param length Length (in pixels) to update
 * @param data Pixels to paint in the drawable (in RGB or RGBA format)
 * @param flags If TTOS_DRAWABLE_HAS_ALPHA, the format will be RGBA; if not, will be RGB
 * @return
 */

uint32_t ttos_update_drawable (struct ttos_object *master, uint32_t answer_port, uint32_t widget_uid, uint32_t offset,
						 uint32_t length, uint8_t *data, uint32_t flags) {

	uint32_t answer_port2;
	uint32_t max_len;

	if ((widget_uid&0xF0000000)!=TTOS_WIDGET_UID_MASK) {
		return (TTOS_ERROR_INVALID_UID);
	}

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;

	max_len=65496/((flags&TTOS_DRAWABLE_HAS_ALPHA) ? 4 : 3);

	while (length>max_len) {
		minibus_start_message(master->conn,TTOS_SM_UPDATE_DRAWABLE,TTOS_NO_ANSWER);
		minibus_write_long(master->conn,widget_uid);
		minibus_write_long(master->conn,offset);
		minibus_write_long(master->conn,flags);
		minibus_write_blob(master->conn,data,65496);
		if (0!=minibus_send(master->conn)) {
			return (TTOS_ERROR_CANT_SEND_DATA);
		}
		offset+=max_len;
		data+=65496;
		length-=max_len;
	}
	minibus_start_message(master->conn,TTOS_SM_UPDATE_DRAWABLE,answer_port2);
	minibus_write_long(master->conn,widget_uid);
	minibus_write_long(master->conn,offset);
	minibus_write_long(master->conn,flags);
	minibus_write_blob(master->conn,data,length*((flags&TTOS_DRAWABLE_HAS_ALPHA) ? 4 : 3));
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}

}

/**
 * Creates a sharemem object, useful for fast-updating drawables
 * @param master The connection to the compositor
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param width Picture width
 * @param height Picture height
 * @param has_alpha If true, the picture will have alpha; if false, it won't
 * @param sharemem A pointer to a pointer where the function will store the new object.
 * @return The sharemem UID, or an Error
 */

uint32_t ttos_new_sharemem(struct ttos_object *master, uint32_t answer_port, uint16_t width, uint16_t height, bool has_alpha,
						   struct ttos_sharemem_object ** sharemem) {

	struct ttos_sharemem_object *object;
	uint32_t uid;
	uint32_t answer_port2;
	int f;
	size_t final_size;
	uint8_t *pixels;
	Cstring shname;
	char tmp[200];

	shname=tmp;
	*sharemem=NULL;

	object=malloc(sizeof(struct ttos_sharemem_object));
	if (!object) {
		return (TTOS_ERROR_NO_MEMORY);
	}

#ifdef USE_ARM
	sprintf(shname,"/dev/shm/%d_%d",master->uid,master->shcounter);
	f=open(shname,O_RDWR | O_CREAT|O_TRUNC, S_IRWXU | S_IRWXG);
#else
	sprintf(shname,"/%d_%d",master->uid,master->shcounter);
	f=shm_open(shname,O_RDWR | O_CREAT|O_TRUNC, S_IRWXU | S_IRWXG);
#endif
	master->shcounter++;
	if (f<0) {
		free(object);
		return TTOS_ERROR_CANT_SHARE_MEM;
	}


	if (has_alpha) {
		final_size=3*width*height;
	} else {
		final_size=2*width*height;
	}

	if (0!=ftruncate(f,final_size)) {
		free(object);
		close(f);
		return TTOS_ERROR_CANT_SHARE_MEM;
	}
	pixels=mmap(NULL,final_size,PROT_READ | PROT_WRITE, MAP_SHARED, f,0);
	close(f);
	if (!pixels) {
		free(object);
		return TTOS_ERROR_CANT_SHARE_MEM;
	}

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_NEW_SHAREMEM,answer_port2);
	minibus_write_short(master->conn,width);
	minibus_write_short(master->conn,height);
	minibus_write_byte(master->conn,has_alpha ? 1 : 0);
	minibus_write_string(master->conn,shname);
	if (0!=minibus_send(master->conn)) {
		munmap(pixels,final_size);
		free(object);
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		uid=0;
	} else {
		uid=ttos_wait_for_error(master,answer_port2);
	}

	if (uid>TTOS_ERROR_NO_ERROR) {
		munmap(pixels,final_size);
		free(object);
		return (uid);
	}

	object->width=width;
	object->height=height;
	object->has_alpha=has_alpha;
	object->uid=uid;
	object->pixel=(void *) pixels;
	object->size=final_size;
	if (has_alpha) {
		object->alpha=pixels+(2*width*height);
	} else {
		object->alpha=NULL;
	}

	*sharemem=object;

	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return (uid);
	}
}

/**
 * Destroys a sharemem object
 *
 * @param master The connection to the compositor
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param sharemem The sharemem object to use
 * @return An error
 */

uint32_t ttos_destroy_sharemem (struct ttos_object *master, uint32_t answer_port, struct ttos_sharemem_object * sharemem) {

	uint32_t answer_port2;
	uint32_t retval;

	if (!master || !sharemem) {
		return TTOS_ERROR_INVALID_OBJECT;
	}

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_DESTROY_SHAREMEM,answer_port2);
	minibus_write_long(master->conn,sharemem->uid);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		retval=ttos_wait_for_error(master,answer_port2);
		if (retval==TTOS_ERROR_NO_ERROR) {
			munmap(sharemem->pixel,sharemem->size);
			free(sharemem);
		}
		return(retval);
	}
}


/**
 * Updates a sharemem object with the pixel data specified, thus updating all drawables linked to it
 *
 * @param master The connection to the compositor
 * @param sharemem The sharemem object to update
 * @param offset Offset (in pixels) to the start point
 * @param length Length (in pixels) to update
 * @param data Pixels to paint in the drawable (in RGB or RGBA format)
 * @param flags If TTOS_DRAWABLE_HAS_ALPHA, the format will be RGBA; if not, will be RGB
 * @return TTOS_ERROR_NO_ERROR if all went fine; other if there was an error
 */

uint32_t ttos_update_sharemem (struct ttos_object *master, struct ttos_sharemem_object * sharemem, uint32_t offset,
								uint32_t length, uint8_t *data, uint32_t flags) {

	bool has_alpha;
	uint32_t p_length,counter;
	uint16_t *pixel_data;
	uint8_t *alpha_data;
	register uint16_t r,g,b,a;

	if (!master || !sharemem) {
		return TTOS_ERROR_INVALID_OBJECT;
	}

	has_alpha = (flags&TTOS_DRAWABLE_HAS_ALPHA) ? true : false;

	p_length=sharemem->width*sharemem->height;
	if (offset>=p_length) {
		return TTOS_ERROR_OFFSET_OUT_OF_RANGE;
	}
	if ((length+offset)>p_length) {
		length=p_length-offset;
	}

	pixel_data=sharemem->pixel+offset;
	if (sharemem->has_alpha) {
		alpha_data=sharemem->alpha+offset;
	} else {
		alpha_data=NULL;
	}

	a=0x00;
	if (has_alpha) {
		if (alpha_data) {
			for(counter=0;counter<length;counter++) {
				r=(uint16_t)(*(data++));
				g=(uint16_t)(*(data++));
				b=(uint16_t)(*(data++));
				a=(uint16_t)(*(data++));
				if (a==0xFF) {
					*(pixel_data++)=((r<<8)&0xF800)|((g<<3)&0x07E0)|((b>>3)&0x001F);
					*(alpha_data++)=0;
					continue;
				}
				if (!a) {
					data+=3;
					*(pixel_data++)=0;
					*(alpha_data++)=0xFF;
					continue;
				}
				*(pixel_data++)=(r&0xF800)|((g>>5)&0x07E0)|((b>>11)&0x001F);
				*(alpha_data++)=~a;
			}
		} else {
			for(counter=0;counter<length;counter++) {
				r=(uint16_t)(*(data++));
				g=(uint16_t)(*(data++));
				b=(uint16_t)(*(data++));
				a=(uint16_t)(*(data++));
				if (a==0xFF) {
					*(pixel_data++)=((r<<8)&0xF800)|((g<<3)&0x07E0)|((b>>3)&0x001F);
					continue;
				}
				if (!a) {
					data+=3;
					*(pixel_data++)=0;
					continue;
				}
				*(pixel_data++)=(r&0xF800)|((g>>5)&0x07E0)|((b>>11)&0x001F);
			}
		}
	} else {
		if (alpha_data) {
			for(counter=0;counter<length;counter++) {
				r=(uint16_t)(*(data++));
				g=(uint16_t)(*(data++));
				b=(uint16_t)(*(data++));
				*(pixel_data++)=((r<<8)&0xF800)|((g<<3)&0x07E0)|((b>>3)&0x001F);
				*(alpha_data++)=0;
			}
		} else {
			for(counter=0;counter<length;counter++) {
				r=(uint16_t)(*(data++));
				g=(uint16_t)(*(data++));
				b=(uint16_t)(*(data++));
				*(pixel_data++)=((r<<8)&0xF800)|((g<<3)&0x07E0)|((b>>3)&0x001F);
			}
		}
	}
	return (TTOS_ERROR_NO_ERROR);
}


/**
 * Registers a new bus in the system
 *
 * @param master The connection to the compositor
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param bus_name The name for the new bus
 * @param flags TTOS_IPC_ flags
 *
 * @return The new bus port, or an Error
 */

uint32_t ttos_register_bus(struct ttos_object *master, uint32_t answer_port, Cstring bus_name, uint32_t flags) {

	uint32_t answer_port2;

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_REGISTER_BUS,answer_port2);
	minibus_write_string(master->conn,bus_name);
	minibus_write_long(master->conn,flags);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Unregisters an existing bus in the system. Only the creator (owner) can do it
 *
 * @param master The connection to the compositor
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param bus_name The bus name
 *
 * @return Error
 */

uint32_t ttos_unregister_bus(struct ttos_object *master, uint32_t answer_port, Cstring bus_name) {

	uint32_t answer_port2;

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_UNREGISTER_BUS,answer_port2);
	minibus_write_string(master->conn,bus_name);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Subscribes the caller to the specified bus, allowing it to send messages (ALL_TO_ONE bus) or receive them (ONE_TO_ALL bus)
 *
 * @param master The connection to the compositor
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param bus_name The bus name
 *
 * @return Error
 */

uint32_t ttos_subscribe_to_bus(struct ttos_object *master, uint32_t answer_port, Cstring bus_name) {

	uint32_t answer_port2;

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_SUBSCRIBE_BUS,answer_port2);
	minibus_write_string(master->conn,bus_name);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}


/**
 * Unsubscribes the caller to the specified bus
 *
 * @param master The connection to the compositor
 * @param answer_port The port into which receive asynchronously the answer, or TTOS_SYNC_ANSWER to wait the answer
 * @param bus_name The bus name
 *
 * @return Error
 */

uint32_t ttos_unsubscribe_to_bus(struct ttos_object *master, uint32_t answer_port, Cstring bus_name) {

	uint32_t answer_port2;

	answer_port2 = (answer_port==TTOS_SYNC_ANSWER) ? TTOS_RESERVED_PORTS+0x02 : answer_port;
	minibus_start_message(master->conn,TTOS_SM_UNSUBSCRIBE_BUS,answer_port2);
	minibus_write_string(master->conn,bus_name);
	if (0!=minibus_send(master->conn)) {
		return (TTOS_ERROR_CANT_SEND_DATA);
	}
	if (answer_port!=TTOS_SYNC_ANSWER) {
		return (TTOS_ERROR_ASYNC_RESPONSE);
	} else {
		return(ttos_wait_for_error(master,answer_port2));
	}
}

