/*
 * Copyright 2011 Raster Software Vigo
 *
 * This file is part of ttos
 *
 * ttos is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * ttos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * ttos_manager.c
 *
 *  Created on: 17/11/2010
 *      Author: raster
 */
#define _BSD_SOURCE
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <signal.h>
#include <libintl.h>

#include "../ttos_lib/ttos.h"
#include "../minibus/minibus.h"
#include "ttos_manager.h"

#define _(string) gettext (string)

struct minibus_master *queue;
struct minibus_list app_list,running_list;
bool visible;
uint32_t current_tab;
bool global_exit;

uint16_t screen_width,screen_height;

/**
 * Creates a new application object
 *
 * @param uid The client's UID
 *
 * @return A new client app object, or NULL on error
 */

struct manager_app* manager_new_app(Cstring program,Cstring icon) {

	struct manager_app *object;

	if ((!program)||(program[0]==0)||(!icon)||(icon[0]==0))
		return NULL;

	object=malloc(sizeof(struct manager_app));
	if (!object) {
		return (NULL);
	}
	object->launch_command=strdup(program);
	object->icon=strdup(icon);
	object->running=NULL;
	return (object);
}

/**
 * Reads a client application template and returns a client app object
 */

struct manager_app *manager_new_app_from_file(Cstring filename) {

	struct manager_app *object;
	int handler;
	Cstring buffer,current,start,program,icon;
	struct stat statbuf;
	off_t size;

	handler=open(filename,O_RDONLY);
	if (handler<0) {
		return NULL;
	}
	if (fstat(handler,&statbuf)<0) {
		close(handler);
		return NULL;
	}
	size=statbuf.st_size;
	buffer=malloc(size+2);
	if (buffer==NULL) {
		close(handler);
		return NULL;
	}
	if (-1==read(handler,buffer,size)) {
		close(handler);
		return NULL;
	}
	close(handler);

	buffer[size]='\n'; // add a newline at end
	buffer[size+1]=0; // transform it in an ASCIIZ string
	program=NULL;
	icon=NULL;
	start=current=buffer;
	while(*current) {
		if (*current<' ') {
			*current=0;
			if (!strncmp(start,"program=",8)) {
				program=start+8;
			} else if (!strncmp(start,"icon=",5)) {
				icon=start+5;
			}
			start=current+1;
		}
		current++;
	}
	if ((!icon)||(!program)) {
		free(buffer);
		return NULL;
	}
	object=manager_new_app(program,icon);
	free(buffer);
	return (object);
}


/**
 * Minibus list callback to delete a client object
 *
 * @param data A pointer to the object to delete
 */

void manager_destroy_app(void *data) {

	struct manager_app *object=data;

	if (object->icon) {
		free(object->icon);
	}
	if (object->launch_command) {
		free(object->launch_command);
	}
	if (object->running) {
		minibus_list_remove_data(&running_list,object,true);
	}
	free(object);
}


/**
 * Creates a new running application object
 *
 * @param uid The client's UID
 * @param program The calling command, as passed in argv[0]
 *
 * @return A new client object, or NULL on error
 */

struct manager_running* manager_new_running(uint32_t uid,Cstring program) {

	struct manager_running *object;

	object=malloc(sizeof(struct manager_running));
	if (!object) {
		return (NULL);
	}
	if ((!program)||(program[0]==0))
		object->program=NULL;
	else
		object->program=strdup(program);
	object->uid=uid;
	object->app=NULL;
	return (object);
}

/**
 * Minibus list callback to delete a client object
 *
 * @param data A pointer to the object to delete
 */

void manager_destroy_running(void *data) {

	struct manager_running *object=data;

	if (object->program) {
		free(object->program);
	}
	free(object);
}


uint32_t manager_scroll_cb(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *p_data) {

	struct ttos_object *master=p_data;

	if (!visible) {
		ttos_destroy_tab(master,TTOS_SYNC_ANSWER,current_tab,0);
		ttos_set_window_visibility(master,TTOS_SYNC_ANSWER,1,false);
		ttos_repaint_screen(master,TTOS_SYNC_ANSWER);
	}

	return 0;
}


void manager_show_main(struct ttos_object *master, bool do_visible) {

	uint32_t old_tab;
	uint16_t counter;
	uint32_t x,y,wx,hy,mx,my,counter2,counter3;
	struct ttos_color color;
	struct minibus_list_element *element;
	struct manager_app *app;

	for(element=app_list.list,counter=0;element!=NULL;element=element->next) {
		app=(struct manager_app *)element->data;
		if (app->running) {
			counter++;
		}
	}

	if (!counter) {
		do_visible=true;
	}

	if (!do_visible) {
		if (visible) {
			visible=false;
			ttos_set_tab_offset(master,TTOS_SYNC_ANSWER,current_tab,-(master->width),0,11,TTOS_TAB_OFFSET_KEEP_VERTICAL|TTOS_TAB_OFFSET_DO_SCROLL);
		}
		return;
	}

	old_tab=current_tab;
	current_tab = (old_tab==1) ? 2 : 1;

	color.r=color.g=color.b=0x3F;
	color.a=0xC0;

	current_tab=ttos_new_tab(master,TTOS_SYNC_ANSWER);
	ttos_new_color_area(master,TTOS_SYNC_ANSWER,current_tab,0,0,master->width,master->height,&color,TTOS_WIDGET_V_FIXED,20);

	wx=master->width/56;
	hy=master->height/56;
	mx=(master->width/(2*wx))-24;
	my=(master->height/(2*hy))-24;
	x=mx-2;
	y=my-2;
	counter2=0;
	counter3=0;
	for(element=app_list.list;element!=NULL;element=element->next) {
		app=(struct manager_app *)element->data;
		if (app->running) {
			color.b=color.g=0xC0;
			color.r=0;
			color.a=0xFF;
		} else {
			color.r=color.g=color.b=0x0F;
			color.a=0xC0;
		}
		app->icon_id=ttos_new_button(master,TTOS_SYNC_ANSWER,current_tab,x,y,52,52,TTOS_NO_STOCK_BUTTON,app->icon,"",NULL,&color,0,10);
		if (app->icon_id==TTOS_ERROR_CANT_LOAD_PICTURE) {
			printf("Can't locate picture %s\n",app->icon);
			app->icon_id=ttos_new_button(master,TTOS_SYNC_ANSWER,current_tab,x,y,52,52,TTOS_NO_STOCK_BUTTON,"noicon.tga","",NULL,&color,0,10);
		}
		x+=2*mx+48;
		counter2++;
		if (counter2>=wx) {
			counter2=0;
			counter3++;
			x=mx-2;
			y+=2*my+48;
		}
	}

	if (counter3<hy) {
		y=((48+my*2)*(hy-1)+my)-2;
	}

	x=((48+mx*2)*(wx-1)+mx)-2;
	ttos_new_button(master,TTOS_SYNC_ANSWER,current_tab,x,y,52,52,TTOS_NO_STOCK_BUTTON,"shutdown.tga","",NULL,NULL,0,12);
	//y+=52;

	if (visible) {
		ttos_set_visible_tab(master,TTOS_SYNC_ANSWER,current_tab,TTOS_TAB_SWAP_MODE_REPLACE, TTOS_NO_ANSWER);
		ttos_destroy_tab(master,TTOS_SYNC_ANSWER,old_tab,0);
	} else {
		ttos_set_tab_offset(master,TTOS_SYNC_ANSWER,current_tab,-(master->width),0,TTOS_NO_ANSWER,0);
		ttos_set_visible_tab(master,TTOS_SYNC_ANSWER,current_tab,TTOS_TAB_SWAP_MODE_REPLACE, TTOS_NO_ANSWER);
		visible=true;
		ttos_set_tab_offset(master,TTOS_SYNC_ANSWER,current_tab,0,0,11,TTOS_TAB_OFFSET_KEEP_VERTICAL|TTOS_TAB_OFFSET_DO_SCROLL);
	}
	ttos_set_window_visibility(master,TTOS_SYNC_ANSWER,1,true);
	ttos_repaint_screen(master,TTOS_SYNC_ANSWER);
}

/**
 * Port 0 callback
 *
 * This callback is called each time an event is received in port 0.
 *
 * @param master The ttos_object which manages the system
 * @param event The event number received
 * @param data Private data
 */


void manager_cb_port0(struct ttos_object *master, uint32_t event, void *data) {

	bool new_visible;

	switch (event) {
	case TTOS_SR_TOUCH_EVENT:
		new_visible=(visible==true) ? false : true;
		manager_show_main(master,new_visible);
	break;
	}
}

/**
 * This callback is called each time a new client connects to the window server
 *
 * @param master The connection object to the window server
 * @param uid The new client's UID
 * @param data Private data for the callback
 */

void manager_new_client(struct ttos_object *master, uint32_t uid,Cstring program, void *data) {

	struct manager_running *client;
	struct minibus_list_element *element;
	struct manager_app *app;

	client=manager_new_running(uid,program);
	if (!client) {
		return;
	}
	minibus_list_add(&running_list,client);
	app=NULL;
	for (element=app_list.list;element!=NULL;element=element->next) {
		app=(struct manager_app *)element->data;
		if (!strcmp(app->launch_command,program)) {
			app->running=client;
			client->app=app;
			break;
		}
	}
	ttos_set_fg_app(master,TTOS_SYNC_ANSWER,uid);
	manager_show_main(master,false);
	ttos_repaint_screen(master,TTOS_SYNC_ANSWER);
}

/**
 * This callback is called each time a client disconnects from the window server
 *
 * @param master The connection object to the window server
 * @param uid The new client's UID
 * @param data Private data for the callback
 */

void manager_destroy_client(struct ttos_object *master, uint32_t uid, uint32_t window, bool was_fg, void *data) {

	struct manager_running *client;
	struct minibus_list_element *element;

	//printf("Global exit: %d running %d %d\n",global_exit,running_list.list,uid);

	for(element=running_list.list;element!=NULL;element=element->next) {
		client=element->data;
		if (client->uid==uid) {
			break;
		}
	}

	if (element) {
		if ((client)&&(client->app)) {
			client->app->running=NULL;
		}
		minibus_list_remove(element,true);
	}

	if(!global_exit) {
		if (was_fg) {
			if (running_list.list) {
				client=running_list.list->data;
				if (client) {
					ttos_set_fg_app(master,TTOS_SYNC_ANSWER,client->uid);
				}
			}
			ttos_repaint_screen(master,TTOS_SYNC_ANSWER);
		}
		manager_show_main(master,visible);
	}

	if ((global_exit)&&(!running_list.list)) {
		minibus_exit_loop(master->conn->master_queue);
	}
}

uint32_t cb_closed_connection(struct minibus_connection *connection,void *data) {

	exit(0);
}

int manager_fill_apps() {

	DIR *directory;
	struct dirent *entry;
	Cstring fname,full_path;
	struct manager_app *app;

	directory=opendir(TTOS_MANAGER_PROGRAM_FOLDER);
	if (directory==NULL) {
		return -1;
	}
	while(NULL!=(entry=readdir(directory))) {
		fname=entry->d_name;
		if ((!strcmp(fname,"."))||(!strcmp(fname,".."))) {
			continue;
		}

		if(TTOS_FTYPE_APP!=ttos_check_extension(fname)) {
			continue;
		}
		full_path=minibus_path_join(TTOS_MANAGER_PROGRAM_FOLDER,fname);
		app=manager_new_app_from_file(full_path);
		free(full_path);
		if(app) {
			minibus_list_add(&app_list,app);
		}
	}
	closedir(directory);
	return 0;
}


uint32_t manager_shutdown_cb(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *p_data) {

	struct manager_running *client;
	struct minibus_list_element *element;
	struct ttos_object *master=p_data;
	struct ttos_dialog_button buttons[3];
	int32_t retval;

	buttons[0].button=TTOS_CANCEL_BUTTON;
	buttons[0].return_value=0;
	buttons[1].button=TTOS_NO_STOCK_BUTTON;
	buttons[1].text=_("Shutdown");
	buttons[1].icon="turnoff.tga";
	buttons[1].return_value=1;

	retval=manager_show_dialog(master,_("Choose action"),TTOS_WIDGET_FONT_MULTILINE|TTOS_WIDGET_FONT_WORD_CUT|TTOS_WIDGET_FONT_ALIGN_CENTER,buttons,2);
	if (retval==0) {
		return 0;
	}

	global_exit=true;

	if (!running_list.list) {
		minibus_exit_loop(connection->master_queue);
		return 0;
	}

	for(element=running_list.list;element!=NULL;element=element->next) {
		client=element->data;
		ttos_close_client(master,TTOS_SYNC_ANSWER,client->uid);
	}
	return 0;
}


uint32_t manager_launch_cb(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *p_data) {

	struct minibus_list_element *element;
	struct minibus_data_element *data;
	struct manager_app *app;
	struct ttos_object *master=p_data;
	int32_t error;
	char *p1[2];

	data=minibus_read_params(connection,"lblllll",&error);

	if(!data) {
		return -1;
	}

	for(element=app_list.list;element!=NULL;element=element->next) {
		app=element->data;
		if (app->icon_id==data[2].value32) {
			break;
		}
	}
	free(data);

	if (element) {
		if (app->running) {
			ttos_set_fg_app(master,TTOS_SYNC_ANSWER,app->running->uid);
			manager_show_main(master,false);
			ttos_repaint_screen(master,TTOS_SYNC_ANSWER);
		} else {
			p1[0]=strdup(app->launch_command);
			p1[1]=NULL;
			ttos_launch_child(p1,NULL);
			free(p1[0]);
		}
	}

	return 0;
}

void kill_handler(int value) {

	int pid;

	do {
		pid = waitpid(-1,NULL,WNOHANG);
	} while (pid>0);
}


int32_t manager_show_dialog(struct ttos_object *master, Cstring text, uint32_t text_flags, struct ttos_dialog_button *buttons, uint16_t count) {

	struct ttos_window_geometry windows[2];
	int32_t retval;

	windows[0].x=0;
	windows[0].y=0;
	windows[0].w=screen_width;
	windows[0].h=screen_height;
	windows[0].visible=true;

	windows[1].x=0;
	windows[1].y=0;
	windows[1].w=screen_width;
	windows[1].h=screen_height;
	windows[1].visible=true;

	ttos_define_windows(master,TTOS_SYNC_ANSWER,2,windows);
	//ttos_set_client_window(master,TTOS_SYNC_ANSWER,master->uid,1);

	retval=ttos_show_dialog(master,text,text_flags,buttons,count);

	ttos_set_visible_tab(master,TTOS_SYNC_ANSWER,current_tab,TTOS_TAB_SWAP_MODE_REPLACE, TTOS_NO_ANSWER);
	ttos_repaint_screen(master,TTOS_SYNC_ANSWER);

	windows[0].x=0;
	windows[0].y=0;
	windows[0].w=screen_width;
	windows[0].h=screen_height;
	windows[0].visible=true;

	windows[1].x=0;
	windows[1].y=0;
	windows[1].w=screen_width;
	windows[1].h=screen_height;
	windows[1].visible=visible;

	ttos_define_windows(master,TTOS_SYNC_ANSWER,2,windows);
	//ttos_set_client_window(master,TTOS_SYNC_ANSWER,master->uid,1);

	return (retval);
}

int main(int argc,char **argv) {

	struct sigaction sg;

	struct ttos_object *master;
	struct ttos_window_geometry windows[2];

	current_tab=2;

	global_exit=false;

	memset(&sg,0,sizeof(struct sigaction));
	sg.sa_handler=kill_handler;
	sigaction(SIGCHLD,&sg,NULL);

	minibus_list_init(&app_list,manager_destroy_app,NULL);
	minibus_list_init(&running_list,manager_destroy_running,NULL);

	master=ttos_init_system (TTOS_INIT_ASK_WM,argc,argv);
	if (!master) {
		printf("Can't init system. Aborting.\n");
		return -1;
	}
	master->cb_port0=manager_cb_port0;
	master->new_client=manager_new_client;
	master->destroyed_client=manager_destroy_client;
	minibus_set_closed_connection_callback(master->queue,cb_closed_connection,NULL);
	minibus_add_callback(master->queue,10,manager_launch_cb,master); // event for application buttons
	minibus_add_callback(master->queue,11,manager_scroll_cb,master); // event for scroll end
	minibus_add_callback(master->queue,12,manager_shutdown_cb,master); // event for shutdown button

	screen_width=master->width;
	screen_height=master->height;

	windows[0].x=0;
	windows[0].y=0;
	windows[0].w=screen_width;
	windows[0].h=screen_height;
	windows[0].visible=true;

	windows[1].x=0;
	windows[1].y=0;
	windows[1].w=screen_width;
	windows[1].h=screen_height;
	windows[1].visible=false;
	visible=windows[1].visible;

	ttos_define_windows(master,TTOS_SYNC_ANSWER,2,windows);
	ttos_set_client_window(master,TTOS_SYNC_ANSWER,master->uid,1);
	manager_fill_apps();

	manager_show_main(master,visible);
	minibus_run_loop(master->queue);

	return 0;

}
