/*
 * Copyright 2011 Raster Software Vigo
 *
 * This file is part of ttos
 *
 * ttos is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * ttos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * uid.c
 *
 *  Created on: 19/12/2010
 *      Author: raster
 */

#include <stdlib.h>
#include <string.h>
#include "ttos_lib/ttos_structs.h"

#include "uid.h"


/**
 * Creates an UID object
 *
 * @return A ttos_uid_c object, or NULL on error
 */

struct ttos_uid_c *uid_init() {

	struct ttos_uid_c *object;

	object=malloc(sizeof(struct ttos_uid_c));
	if (!object) {
		return NULL;
	}
	memset(object->bitmask,0xFF,TTOS_UID_SIZE);
	return (object);
}

/**
 * Destroys an UID object
 *
 * @param object The UID object to destroy
 *
 */

void uid_destroy(struct ttos_uid_c *object) {

	if (!object) {
		return;
	}
	free(object);
}

/**
 * Returns an UID with the specified type
 *
 * @param object The UID object from which obtain the UID
 * @param type The type to assign to the UID
 *
 * @return an UID of the desired type, or 0 if there are no more UIDs available
 */

uint32_t uid_get_uid(struct ttos_uid_c *object, uint32_t type) {

	uint32_t loop,loop2;
	uint8_t mask;
	uint8_t *p;

	p=object->bitmask;
	for(loop=0;loop<TTOS_UID_SIZE;loop++) {
		if (*p) { // here is, at least, one free UID
			mask=0x01;
			loop<<=3;
			for(loop2=0;loop2<8;loop2++) {
				if((*p)&mask) {
					*p^=mask; // set the bit to 0
					return(type|(loop+loop2));
				}
				mask<<=1;
			}
		}
		p++;
	}
	return 0;
}


/**
 * Marks an UID as available, when its assigned resource is freed
 *
 * @param object The UID object from which that UID was obtained
 * @param uid The UID to retrieve
 */


void uid_release_uid(struct ttos_uid_c *object, uint32_t uid) {

	uint32_t pos,mask;

	pos=(uid&0x0FFF)>>3;

	mask=(0x00000001)<<(uid&0x07);
	object->bitmask[pos]|=mask; // set the bit to 1

}


/**
 * Generates a composed UID, based on the widget base UID, the tab's UID and the type
 *
 * @param uid The widget's UID
 * @param tab_uid The tab's UID
 * @param type_mask The mask to add to the UID (0x10000000 for a widget UID, 0x20000000 for a tab UID, 0x30000000 for a timer UID)
 */

uint32_t uid_mix_uid(uint16_t w_uid, uint16_t t_uid, uint32_t type_mask) {

	uint32_t c_uid;

	if (type_mask==TTOS_WIDGET_UID_MASK) {
		c_uid=((uint32_t)w_uid) | ((((uint32_t)t_uid)<<16)&0xFFFF0000);
		c_uid&=0x0FFFFFFF;
	} else {
		c_uid=(uint32_t)w_uid;
		c_uid&=0x0000FFFF;
	}
	c_uid|=type_mask;

	return (c_uid);
}
